# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from joulescope.usb import core as usb_core
from joulescope.usb.api import DeviceEvent
from joulescope.usb.impl_tools import RunUntilDone
from joulescope.usb.scan_info import INFO
from .setupapi import device_interface_guid_to_paths
from ctypes import windll, Structure, POINTER, byref, sizeof, \
    c_ubyte, c_ushort, c_ulong, c_void_p, pointer
from ctypes.wintypes import DWORD, HANDLE, BOOL
import numpy as np
from typing import List
import time

from . import kernel32
import logging
log = logging.getLogger(__name__)

_winusb = windll.winusb

TICK_INTERVAL = 1.0  # seconds
BULK_IN_LENGTH = 512

# Set the control timeout in milliseconds
# WinUSB cannot cancel control transfers.
# In order to prevent heap corruption, the timeout must be less than the
# any forced closing time.
CONTROL_TIMEOUT = 1.0  # seconds


class PipeInfo(Structure):
    _fields_ = [
        ('pipe_type', c_ulong),
        ('pipe_id', c_ubyte),
        ('maximum_packet_size', c_ushort), 
        ('interval', c_ubyte)
    ]

    def __repr__(self):
        return usb_core.structure_to_repr(self)


# https://docs.microsoft.com/en-us/windows-hardware/drivers/usbcon/winusb-functions-for-pipe-policy-modification
class PipePolicy:
    SHORT_PACKET_TERMINATE = 0x01
    AUTO_CLEAR_STALL = 0x02
    PIPE_TRANSFER_TIMEOUT = 0x03
    IGNORE_SHORT_PACKETS = 0x04
    ALLOW_PARTIAL_READS = 0x05
    AUTO_FLUSH = 0x06
    RAW_IO = 0x07
    MAXIMUM_TRANSFER_SIZE = 0x08
    RESET_PIPE_ON_RESUME = 0x09


# BOOL __stdcall WinUsb_Initialize(
#   _In_  HANDLE                   DeviceHandle,
#   _Out_ PWINUSB_INTERFACE_HANDLE InterfaceHandle);
WinUsb_Initialize = _winusb.WinUsb_Initialize
WinUsb_Initialize.restype = BOOL
WinUsb_Initialize.argtypes = [HANDLE, POINTER(c_void_p)]

# BOOL __stdcall WinUsb_Free(
#   _In_ WINUSB_INTERFACE_HANDLE InterfaceHandle);
WinUsb_Free = _winusb.WinUsb_Free
WinUsb_Free.restype = BOOL
WinUsb_Free.argtypes = [c_void_p]

# BOOL __stdcall WinUsb_GetDescriptor(
#   _In_  WINUSB_INTERFACE_HANDLE InterfaceHandle,
#   _In_  UCHAR                   DescriptorType,
#   _In_  UCHAR                   Index,
#   _In_  USHORT                  LanguageID,
#   _Out_ PUCHAR                  Buffer,
#   _In_  ULONG                   BufferLength,
#   _Out_ PULONG                  LengthTransferred);
WinUsb_GetDescriptor = _winusb.WinUsb_GetDescriptor
WinUsb_GetDescriptor.restype = BOOL
WinUsb_GetDescriptor.argtypes = [c_void_p, c_ubyte, c_ubyte, c_ushort, POINTER(c_ubyte), c_ulong, POINTER(c_ulong)]


# BOOL __stdcall WinUsb_QueryDeviceInformation(
#   _In_    WINUSB_INTERFACE_HANDLE InterfaceHandle,
#   _In_    ULONG                   InformationType,
#   _Inout_ PULONG                  BufferLength,
#   _Out_   PVOID                   Buffer);
WinUsb_QueryDeviceInformation = _winusb.WinUsb_QueryDeviceInformation
WinUsb_QueryDeviceInformation.restype = BOOL
WinUsb_QueryDeviceInformation.argtypes = [c_void_p, c_ulong, POINTER(c_ulong), c_void_p]

# BOOL __stdcall WinUsb_QueryInterfaceSettings(
#   _In_  WINUSB_INTERFACE_HANDLE   InterfaceHandle,
#   _In_  UCHAR                     AlternateSettingNumber,
#   _Out_ PUSB_INTERFACE_DESCRIPTOR UsbAltInterfaceDescriptor);
WinUsb_QueryInterfaceSettings = _winusb.WinUsb_QueryInterfaceSettings
WinUsb_QueryInterfaceSettings.restype = BOOL
WinUsb_QueryInterfaceSettings.argtypes = [c_void_p, c_ubyte, POINTER(usb_core.InterfaceDescriptor)]

# BOOL __stdcall WinUsb_GetAssociatedInterface(
#   _In_  WINUSB_INTERFACE_HANDLE  InterfaceHandle,
#   _In_  UCHAR                    AssociatedInterfaceIndex,
#   _Out_ PWINUSB_INTERFACE_HANDLE AssociatedInterfaceHandle); -- must call WinUsb_Free
WinUsb_GetAssociatedInterface = _winusb.WinUsb_GetAssociatedInterface
WinUsb_GetAssociatedInterface.restype = BOOL
WinUsb_GetAssociatedInterface.argtypes = [c_void_p, c_ubyte, POINTER(c_void_p)]

# BOOL __stdcall WinUsb_ControlTransfer(
#   _In_      WINUSB_INTERFACE_HANDLE InterfaceHandle,
#   _In_      WINUSB_SETUP_PACKET     SetupPacket,
#   _Out_     PUCHAR                  Buffer,
#   _In_      ULONG                   BufferLength,
#   _Out_opt_ PULONG                  LengthTransferred,
#   _In_opt_  LPOVERLAPPED            Overlapped);
WinUsb_ControlTransfer = _winusb.WinUsb_ControlTransfer
WinUsb_ControlTransfer.restype = BOOL
WinUsb_ControlTransfer.argtypes = [c_void_p, usb_core.SetupPacket, POINTER(c_ubyte), c_ulong, POINTER(c_ulong), POINTER(
    kernel32.Overlapped)]

# BOOL __stdcall WinUsb_QueryPipe(
#   _In_  WINUSB_INTERFACE_HANDLE  InterfaceHandle,
#   _In_  UCHAR                    AlternateInterfaceNumber,
#   _In_  UCHAR                    PipeIndex,
#   _Out_ PWINUSB_PIPE_INFORMATION PipeInformation);
WinUsb_QueryPipe = _winusb.WinUsb_QueryPipe
WinUsb_QueryPipe.restype = BOOL
WinUsb_QueryPipe.argtypes = [c_void_p, c_ubyte, c_ubyte, POINTER(PipeInfo)]

# BOOL __stdcall WinUsb_ReadPipe(
#   _In_      WINUSB_INTERFACE_HANDLE InterfaceHandle,
#   _In_      UCHAR                   PipeID,
#   _Out_     PUCHAR                  Buffer,
#   _In_      ULONG                   BufferLength,
#   _Out_opt_ PULONG                  LengthTransferred,
#   _In_opt_  LPOVERLAPPED            Overlapped);
WinUsb_ReadPipe = _winusb.WinUsb_ReadPipe
WinUsb_ReadPipe.restype = BOOL
WinUsb_ReadPipe.argtypes = [c_void_p, c_ubyte, POINTER(c_ubyte), c_ulong, POINTER(c_ulong),
                            POINTER(kernel32.Overlapped)]

# BOOL WinUsb_SetPipePolicy(
#   WINUSB_INTERFACE_HANDLE InterfaceHandle,
#   UCHAR                   PipeID,
#   ULONG                   PolicyType,
#   ULONG                   ValueLength,
#   PVOID                   Value);
WinUsb_SetPipePolicy = _winusb.WinUsb_SetPipePolicy
WinUsb_SetPipePolicy.restype = BOOL
WinUsb_SetPipePolicy.argtypes = [c_void_p, c_ubyte, c_ulong, c_ulong, c_void_p]

# BOOL __stdcall WinUsb_WritePipe(
#   _In_      WINUSB_INTERFACE_HANDLE InterfaceHandle,
#   _In_      UCHAR                   PipeID,
#   _In_      PUCHAR                  Buffer,
#   _In_      ULONG                   BufferLength,
#   _Out_opt_ PULONG                  LengthTransferred,
#   _In_opt_  LPOVERLAPPED            Overlapped);
WinUsb_WritePipe = _winusb.WinUsb_WritePipe
WinUsb_WritePipe.restype = BOOL
WinUsb_WritePipe.argtypes = [c_void_p, c_ubyte, POINTER(c_ubyte), c_ulong, POINTER(c_ulong),
                             POINTER(kernel32.Overlapped)]

# BOOL __stdcall WinUsb_AbortPipe(
#   _In_ WINUSB_INTERFACE_HANDLE InterfaceHandle,
#   _In_ UCHAR                   PipeID);
WinUsb_AbortPipe = _winusb.WinUsb_AbortPipe
WinUsb_AbortPipe.restype = BOOL
WinUsb_AbortPipe.argtypes = [c_void_p, c_ubyte]

# BOOL __stdcall WinUsb_FlushPipe(
#   _In_ WINUSB_INTERFACE_HANDLE InterfaceHandle,
#   _In_ UCHAR                   PipeID);
WinUsb_FlushPipe = _winusb.WinUsb_FlushPipe
WinUsb_FlushPipe.restype = BOOL
WinUsb_FlushPipe.argtypes = [c_void_p, c_ubyte]

# BOOL __stdcall WinUsb_GetOverlappedResult(
#   _In_  WINUSB_INTERFACE_HANDLE InterfaceHandle,
#   _In_  LPOVERLAPPED            lpOverlapped,
#   _Out_ LPDWORD                 lpNumberOfBytesTransferred,
#   _In_  BOOL                    bWait);
WinUsb_GetOverlappedResult = _winusb.WinUsb_GetOverlappedResult
WinUsb_GetOverlappedResult.restype = BOOL
WinUsb_GetOverlappedResult.argtypes = [c_void_p, POINTER(kernel32.Overlapped), POINTER(DWORD), BOOL]


def wrap_no_throw(f):
    if f is None:
        return lambda *args, **kwargs: None

    def fn(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception:
            log.exception('While calling %s' % fn.__name__)
    return fn


def sanitize_boolean_return_code(result):
    if result:
        return 0
    else:
        return kernel32.GetLastError()


class TransferOverlapped:
    def __init__(self, event, size):
        self.ov = kernel32.Overlapped(event)
        self.ptr = pointer(self.ov)
        self.data = np.full(size, 0, dtype=np.uint8)
        self.b = self.data.ctypes.data_as(POINTER(c_ubyte))
        self.size = size

    def reset(self):
        self.ov.reset()


class EndpointIn:

    def __init__(self, winusb, pipe_id, transfers, block_size, data_fn, process_fn, stop_fn):
        """Manage an in endpoint.

        :param winusb: The underlying winusb handle.
        :param pipe_id: The endpoint IN pipe identifier.
        :param transfers: The number of outstanding transfers to pend.
        :param block_size: The size of each transfer in bytes.
        :param data_fn: The function to call with the received endpoint IN data.
            The data_fn should normally return False, but can return True to
            stop the endpoint streaming.
        :param process_fn: The function() called after data_fn was called.
            This function can have more latency than data_fn.
            The process_fn should normally return False, but can return True to
            stop the endpoint streaming.
        :param stop_fn: The function(event, message) called when this endpoint
            stops streaming data.
        """
        self._winusb = winusb
        self.pipe_id = pipe_id  # (endpoint_id & 0x7f) | 0x80
        self._overlapped_free = []
        self._overlapped_pending = []
        self._transfers = transfers
        self._transfer_size = ((block_size + BULK_IN_LENGTH - 1) // BULK_IN_LENGTH) * BULK_IN_LENGTH
        self._data_fn = data_fn
        self._process_fn = process_fn
        self._stop_fn = stop_fn
        self._process_transfers = 0
        self._time_last = 0.0

        self._state = self.ST_IDLE
        self.stop_code = None
        self.stop_message = ''
        self.byte_count_this = 0
        self.byte_count_total = 0
        self.transfer_count = 0
        self.transfer_expire_max = 0

        self.event = None
        self._overlapped_free = []
        self._overlapped_pending = []

    ST_IDLE = 0
    ST_RUNNING = 1
    ST_STOPPING = 2

    def _open(self):
        self.stop_code = None
        self.event = kernel32.CreateEvent(
            None,   # default security
            True,   # manual reset
            False,  # not initially signalled
            None    # no name
        )
        if self.event is None:
            raise ValueError('could not create event')

        for i in range(self._transfers):
            ov = TransferOverlapped(self.event, self._transfer_size)
            self._overlapped_free.append(ov)

    def _close(self):
        if self.event is not None:
            kernel32.CloseHandle(self.event)
        self.event = None

    def _issue(self, ov):
        ov.reset()
        result = WinUsb_ReadPipe(self._winusb, self.pipe_id, ov.b, ov.size, None, ov.ptr)
        ec = kernel32.GetLastError()
        if not result and (ec != kernel32.ERROR_IO_PENDING):
            msg = 'EndpointIn %d issue failed: %s' % (self.pipe_id, kernel32.get_error_str(ec))
            self._overlapped_free.append(ov)
            self._halt(DeviceEvent.COMMUNICATION_ERROR, msg)
            return True
        self._overlapped_pending.append(ov)
        return False

    def _pend(self):
        while len(self._overlapped_free):
            ov = self._overlapped_free.pop(0)
            if self._issue(ov):
                return True
        return False

    def _expire(self):
        rv = False
        length_transferred = c_ulong(0)
        count = 0
        while not rv and len(self._overlapped_pending):
            ov = self._overlapped_pending[0]
            if WinUsb_GetOverlappedResult(self._winusb, ov.ptr, byref(length_transferred), False):
                ov = self._overlapped_pending.pop(0)
                self.transfer_count += 1
                length = length_transferred.value
                self.byte_count_this += length
                count += 1
                try:
                    rv = self._data_fn(ov.data[:length])
                    rv = bool(rv)
                except Exception:
                    log.exception('data_fn exception: stop streaming')
                    rv = True
                if rv:
                    msg = f'EndpointIn {self.pipe_id} terminated by data_fn'
                    self._halt(DeviceEvent.ENDPOINT_CALLBACK_STOP, msg)
                    self._overlapped_free.append(ov)
                else:
                    rv = self._issue(ov)
            else:
                ec = kernel32.GetLastError()
                if ec in [kernel32.ERROR_IO_INCOMPLETE, kernel32.ERROR_IO_PENDING]:
                    break
                ov = self._overlapped_pending.pop(0)
                self._overlapped_free.append(ov)
                msg = 'EndpointIn WinUsb_GetOverlappedResult fatal: %s' % (kernel32.get_error_str(ec), )
                rv = True
                self._halt(DeviceEvent.COMMUNICATION_ERROR, msg)
        if count > self.transfer_expire_max:
            self.transfer_expire_max = count
        self._process_transfers += count
        return rv

    def _cancel(self):
        length_transferred = c_ulong(0)
        if not WinUsb_AbortPipe(self._winusb, self.pipe_id):
            log.warning('WinUsb_AbortPipe pipe_id %d: %s', self.pipe_id, kernel32.get_last_error())
        while len(self._overlapped_pending):
            ov = self._overlapped_pending.pop(0)
            if not WinUsb_GetOverlappedResult(self._winusb, ov.ptr, byref(length_transferred), True):
                ec = kernel32.GetLastError()
                if ec == kernel32.ERROR_IO_OPERATION_ABORTED:
                    pass  # aborted as requested, no issue
                elif ec in [kernel32.ERROR_FILE_NOT_FOUND, kernel32.ERROR_GEN_FAILURE, kernel32.ERROR_DEVICE_NOT_CONNECTED]:
                    log.debug('cancel overlapped: %s', kernel32.get_last_error())
                else:
                    log.warning('cancel overlapped: %s', kernel32.get_last_error())
            self._overlapped_free.append(ov)

    def _halt(self, stop_code, msg=None):
        if self._state != self.ST_STOPPING:
            self._state = self.ST_STOPPING
            self._cancel()
        if stop_code:
            lvl = logging.INFO if stop_code <= 0 else logging.ERROR
            if self.stop_code is None:
                self.stop_code = stop_code
                self.stop_message = '' if msg is None else str(msg)
                log.log(lvl, 'endpoint halt %d: %s', stop_code, msg)
            else:
                log.log(lvl, 'endpoint halt %d duplicate: %s', stop_code, msg)

    def process(self):
        """Process pending data.

        :return: True to stop process, False to continue processing.
        """
        if self._state != self.ST_RUNNING:
            return self.stop_code > 0
        rv = self._expire()
        if not rv:
            rv = self._pend()
        return rv

    def process_signal(self):  #: NoRaise
        if self._process_transfers:
            self._process_transfers = 0
            try:
                if callable(self._process_fn):
                    return self._process_fn()
            except Exception:
                log.exception('_process_fn exception: stop streaming')
                return True  # force stop
        return False

    def start(self):
        log.info("endpoint start 0x%02x transfer size = %d bytes" % (self.pipe_id, self._transfer_size))
        self._open()
        self._state = self.ST_RUNNING
        self._process_transfers = 0
        self._time_last = time.time()
        self._pend()

    def stop(self):
        if self._state != self.ST_IDLE:
            log.info("endpoint stop")
            if self._state != self.ST_STOPPING:
                self._cancel()  # was already called in halt
            if self.stop_code is None:
                self.stop_code = 0
                self.process_signal()  # ensure that all received data is processed
            self._close()
            try:
                self._stop_fn(self.stop_code, self.stop_message)
            except Exception:
                log.exception('_stop_fn exception')
            self._state = self.ST_IDLE

    def status(self):
        """Get the endpoint status.

        :return: A dict mapping status name to a value..  The value is a dict
            containing 'value' and 'units'.
        """
        time_now = time.time()
        duration = time_now - self._time_last
        if duration < 0.01:
            throughput = 0.0
        else:
            throughput = self.byte_count_this / duration
        self.byte_count_total += self.byte_count_this
        status = {
            'bytes': {'value': self.byte_count_total, 'units': 'bytes'},
            'transfers': {'value': self.transfer_count, 'units': 'transfers'},
            'duration': {'value': duration, 'units': 'seconds'},
            'throughput': {'value': throughput, 'units': 'Bps'},
            'transfer_expire': {'value': self.transfer_expire_max, 'units': 'transfers'},

        }
        self.byte_count_this = 0
        self.transfer_expire_max = 0
        self._time_last = time_now
        return status


class ControlTransferAsync:

    def __init__(self, winusb):
        self._winusb = winusb
        self._overlapped = None
        self._event = None
        self._commands = []  # list of pending control transfer commands
        self._time_start = None
        self.stop_code = None

    @property
    def event(self):
        return self._event

    def open(self):
        self.stop_code = None
        self._event = kernel32.CreateEvent(
            None,   # default security
            True,   # manual reset
            False,  # not initially signalled
            None    # no name
        )
        if self._event is None:
            raise ValueError('could not create control event')
        self._overlapped = TransferOverlapped(self._event, 4096)

    def close(self):
        self._commands, commands = [], self._commands
        # cannot abort control pipe using WinUSB!
        commands_len = len(commands)
        if commands:
            command = commands.pop(0)
            rc = kernel32.WaitForSingleObject(self._event, int(CONTROL_TIMEOUT * 1100))
            if rc != kernel32.WAIT_OBJECT_0:  # transfer done
                log.warning('ControlTransferAsync.close but transaction hung')
            self._finish(command)
        while commands:
            cbk_fn, setup_packet, _ = commands.pop(0)
            cbk_fn(usb_core.ControlTransferResponse(setup_packet, False, None))
        if commands_len == 0:
            if self._event:
                kernel32.CloseHandle(self._event)
                self._event = None
            if self._overlapped:
                self._overlapped = None
        else:
            pass  # defer to _close_event

    def _close_event(self):
        if self._event:
            kernel32.CloseHandle(self._event)
            self._event = None
            if self._overlapped:
                self._overlapped = None

    def pend(self, cbk_fn, setup_packet: usb_core.SetupPacket, buffer=None):
        """Pend an asynchronous Control Transfer.

        :param cbk_fn: The function to call when the control transfer completes.
            A :class:`usb_core.ControlTransferResponse` is the sole argument.
        :param setup_packet:
        :param buffer: The buffer (if length > 0) for write transactions.
        :return: True on pending, False on error.
        """
        if self.stop_code is not None:
            response = usb_core.ControlTransferResponse(setup_packet, self.stop_code, None)
            cbk_fn(response)
            return False
        command = [cbk_fn, setup_packet, buffer]
        was_empty = not bool(self._commands)
        self._commands.append(command)
        if was_empty:
            return self._issue()
        return True

    def _issue(self):
        if not self._commands:
            return True
        cbk_fn, setup_packet, buffer = self._commands[0]
        pkt = usb_core.RequestType(value=setup_packet.request_type)
        self._overlapped.reset()
        if pkt.direction == 'out':
            if setup_packet.length > 0:
                log.debug('ControlTransferAsync._issue buffer type: %s', type(buffer))
                self._overlapped.data[:setup_packet.length] = np.frombuffer(buffer, dtype=np.uint8)
        result = WinUsb_ControlTransfer(
            self._winusb,
            setup_packet,
            self._overlapped.b,
            setup_packet.length,
            None,
            self._overlapped.ptr)
        result = sanitize_boolean_return_code(result)
        self._time_start = time.time()
        if result != kernel32.ERROR_IO_PENDING:
            if self.stop_code is None:
                self.stop_code = DeviceEvent.COMMUNICATION_ERROR
            log.warning('ControlTransferAsync._issue %s', kernel32.get_error_str(result))
            response = usb_core.ControlTransferResponse(setup_packet, result, None)
            cbk_fn(response)
            return False
        return True

    def _finish(self, command):
        buffer = None
        cbk_fn, setup_packet, _ = command
        length_transferred = c_ulong()
        rc = WinUsb_GetOverlappedResult(self._winusb, self._overlapped.ptr, byref(length_transferred), True)
        if rc == 0:  # failed (any value other than 0 is success)
            rc = kernel32.GetLastError()
            if rc not in [kernel32.ERROR_IO_INCOMPLETE, kernel32.ERROR_IO_PENDING]:
                kernel32.ResetEvent(self._event)
                log.warning('ControlTransferAsync._finish result: %s', kernel32.get_error_str(rc))
        else:
            kernel32.ResetEvent(self._event)
            rc = 0
            pkt = usb_core.RequestType(value=setup_packet.request_type)
            duration = time.time() - self._time_start
            if pkt.direction == 'in' and setup_packet.length:
                log.debug('ControlTransferAsync._finish duration=%.6f s, length: %s, %s',
                          duration, setup_packet.length, length_transferred.value)
                buffer = bytes(self._overlapped.data[:length_transferred.value])
            else:
                log.debug('ControlTransferAsync._finish duration=%.6f s', duration)
        response = usb_core.ControlTransferResponse(setup_packet, rc, buffer)
        cbk_fn(response)

    def process(self):
        """Check if control transfer completed"""
        if not self._commands:
            return
        rc = kernel32.WaitForSingleObject(self._event, 0)
        if rc == kernel32.WAIT_OBJECT_0:  # transfer done
            command = self._commands.pop(0)
            self._finish(command)
            if self.stop_code is None:
                self._issue()
            else:
                self._close_event()
        elif rc == kernel32.WAIT_TIMEOUT:
            pass  # not ready yet
        else:
            log.warning('ControlTransferAsync.process wait %d', rc)


class WinUsbDevice:
    """The WinUSB :class:`usb.api.Device` implementation"""

    def __init__(self, path):
        self._path = path
        self._file = None  # The file handle for this device
        self._winusb = HANDLE()  # The WinUSB handle for the device's default interface
        self._interface = 0
        self._endpoints = {}  # type: Dict[int, EndpointIn]
        self._event = None
        self._event_list = (HANDLE * kernel32.MAXIMUM_WAIT_OBJECTS)()
        self._event_callback_fn = None
        self._control_transfer = None
        self._update_event_list()

    def __str__(self):
        # path like: \\?\usb#vid_1fc9&pid_fc93#0001#{99a06894-3518-41a5-a207-8519746da89f}
        # not recommended by Microsoft, but good enough for now (20181003)
        try:
            parts = self._path.split('#')
            # vid, pid = [x.split('_')[1] for x in parts[1].split('&')]
            serial_number = parts[2]
            return 'Joulescope:' + serial_number
        except Exception:
            return 'Joulescope'

    def _update_event_list(self):
        self.event_list_count = 0

        def _event_append(event):
            if event is not None:
                self._event_list[self.event_list_count] = event
                self.event_list_count += 1

        _event_append(self._event)
        if self._control_transfer:
            _event_append(self._control_transfer.event)
        for endpoint in self._endpoints.values():
            _event_append(endpoint.event)

    @property
    def path(self):
        return self._path

    @property
    def serial_number(self):
        parts = self._path.split('#')
        if len(parts) != 4:
            return '0000-0000'
        return parts[2]

    def open(self, event_callback_fn=None):
        self.close()
        log.info('WinUsbDevice.open')
        self._event_callback_fn = event_callback_fn

        # create event for self.signal()
        self._event = kernel32.CreateEvent(
            None,   # default security
            True,   # manual reset
            False,  # not initially signalled
            None    # no name
        )

        try:
            if self._event is None:
                raise RuntimeError('could not create event')

            self._file = kernel32.CreateFile(
                self._path,
                kernel32.GENERIC_WRITE | kernel32.GENERIC_READ,
                kernel32.FILE_SHARE_WRITE | kernel32.FILE_SHARE_READ,
                None,
                kernel32.OPEN_EXISTING,
                kernel32.FILE_ATTRIBUTE_NORMAL | kernel32.FILE_FLAG_OVERLAPPED,
                None)

            if self._file == kernel32.INVALID_HANDLE_VALUE:
                raise IOError('%s : open failed on invalid handle' % self)
            result = WinUsb_Initialize(self._file, byref(self._winusb))
            if result == 0:
                raise IOError('%s : open failed %s' % (self, kernel32.get_last_error()))
            log.info('is_high_speed = %s', self._is_high_speed)
            log.info('interface_settings = %s', self._query_interface_settings(0))
            self._control_transfer = ControlTransferAsync(self._winusb)
            self._control_transfer.open()
            timeout = DWORD(int(CONTROL_TIMEOUT * 1000))
            result = WinUsb_SetPipePolicy(self._winusb, 0, PipePolicy.PIPE_TRANSFER_TIMEOUT,
                                          sizeof(timeout), byref(timeout))
            if not result:
                log.warning('WinUsb_SetPipePolicy: %s', kernel32.get_error_str())
            self._update_event_list()
        except Exception:
            self.close()
            raise

    def close(self):
        log.info('WinUsbDevice.close')
        for endpoint in self._endpoints.values():
            endpoint.stop()
        self._endpoints.clear()
        if self._control_transfer is not None:
            self._control_transfer.close()
            self._control_transfer = None
        if self._file is not None:
            WinUsb_Free(self._winusb)
            self._winusb = HANDLE()
            kernel32.CloseHandle(self._file)
            self._file = None
        self._interface = 0
        if self._event:
            kernel32.CloseHandle(self._event)
            self._event = None
        self._event_callback_fn = None

    @property
    def _is_high_speed(self):
        if self._file is None:
            raise IOError('WinUSB device not open')
        buff = (c_void_p * 1)()
        buff_length = c_ulong(sizeof(buff))
        result = WinUsb_QueryDeviceInformation(self._winusb, 1, byref(buff_length), buff)
        if result != 0:
            return buff[0] >= 3
        else:
            raise IOError('WinUsb_QueryDeviceInformation failed')
    
    def _query_interface_settings(self, interface, alternate_setting=0):
        if interface != 0:
            handle = HANDLE()
            result = WinUsb_GetAssociatedInterface(self._winusb, interface, byref(handle))
            if result == 0:
                raise IOError('WinUsb_GetAssociatedInterface failed: ' + kernel32.get_last_error())
        else:
            handle = self._winusb
        interface_descriptor = usb_core.InterfaceDescriptor()
        result = WinUsb_QueryInterfaceSettings(handle, alternate_setting, byref(interface_descriptor))
        if interface != 0:
            WinUsb_Free(handle)
        if result == 0:
            raise IOError('WinUsb_QueryInterfaceSettings failed: ' + kernel32.get_last_error())
        return interface_descriptor

    def _query_pipe(self, pipe_index):
        pipe_info = PipeInfo()
        result = WinUsb_QueryPipe(self._winusb, 0, pipe_index, byref(pipe_info))
        if result == 0:
            raise IOError('WinUsb_QueryPipe failed: ' + kernel32.get_last_error())
        return pipe_info

    def control_transfer_out(self, cbk_fn, recipient, type_, request, value=0, index=0, data=None):
        if cbk_fn is None:
            run_until_done = RunUntilDone(1.0, 'control_transfer_out')
            self.control_transfer_out(run_until_done.cbk_fn, recipient, type_, request, value, index, data)
            while not run_until_done.is_done():
                self.process(0.01)
            return run_until_done.value_args0
        log.debug('control_transfer_out')
        request_type = usb_core.RequestType(direction='out', type_=type_, recipient=recipient).u8
        length = 0 if data is None else len(data)
        pkt = usb_core.SetupPacket(request_type, request, value, index, length)
        return self._control_transfer.pend(cbk_fn, pkt, data)

    def control_transfer_in(self, cbk_fn, recipient, type_, request, value, index, length):
        if cbk_fn is None:
            run_until_done = RunUntilDone(1.0, 'control_transfer_in')
            self.control_transfer_in(run_until_done.cbk_fn, recipient, type_, request, value, index, length)
            while not run_until_done.is_done():
                self.process(0.01)
            return run_until_done.value_args0
        log.debug('control_transfer_in')
        request_type = usb_core.RequestType(direction='in', type_=type_, recipient=recipient).u8
        pkt = usb_core.SetupPacket(request_type, request, value, index, length)
        return self._control_transfer.pend(cbk_fn, pkt)

    def read_stream_start(self, endpoint_id, transfers, block_size, data_fn, process_fn, stop_fn):
        log.info('read_stream_start %d', endpoint_id)
        pipe_id = (endpoint_id & 0x7f) | 0x80
        endpoint = self._endpoints.pop(pipe_id, None)
        if endpoint is not None:
            log.warning('repeated start')
            endpoint.stop()
        endpoint = EndpointIn(self._winusb, pipe_id, transfers,
                              block_size, data_fn, process_fn, stop_fn)
        self._endpoints[endpoint.pipe_id] = endpoint
        endpoint.start()
        self._update_event_list()

    def read_stream_stop(self, endpoint_id):
        log.info('read_stream_stop %d', endpoint_id)
        pipe_id = (endpoint_id & 0x7f) | 0x80
        endpoint = self._endpoints.pop(pipe_id, None)
        if endpoint is not None:
            endpoint.stop()
            self._update_event_list()

    def status(self):
        e = {}
        s = {'endpoints': e}
        for endpoint in self._endpoints.values():
            e[endpoint.pipe_id] = endpoint.status()
        return s

    def signal(self):
        kernel32.SetEvent(self._event)

    def _abort(self, stop_code, msg):
        for endpoint in self._endpoints.values():
            endpoint.stop()
        self._endpoints.clear()
        if self._control_transfer.stop_code is None:
            self._control_transfer.stop_code = DeviceEvent.ENDPOINT_CALLBACK_STOP
        self._update_event_list()
        event_callback_fn, self._event_callback_fn = self._event_callback_fn, None
        if callable(event_callback_fn):
            try:
                event_callback_fn(stop_code, msg)
            except Exception:
                log.exception('while in _event_callback_fn')

    def process(self, timeout):
        timeout_ms = int(timeout * 1000)
        rv = kernel32.WaitForMultipleObjects(self.event_list_count, self._event_list, False, timeout_ms)
        if rv < kernel32.MAXIMUM_WAIT_OBJECTS:
            kernel32.ResetEvent(self._event)
            stop_endpoint_ids = []
            for endpoint in self._endpoints.values():
                if endpoint.process():
                    stop_endpoint_ids.append(endpoint.pipe_id)
            for endpoint in self._endpoints.values():
                if endpoint.process_signal() or endpoint.stop_code is not None:
                    stop_endpoint_ids.append(endpoint.pipe_id)
            pipe_id = None
            for pipe_id in stop_endpoint_ids:
                endpoint = self._endpoints.pop(pipe_id, None)
                if endpoint is None:
                    continue  # duplicate
                endpoint.stop()
                self._update_event_list()
                msg = 'Endpoint pipe_id %r stopped: %s' % (pipe_id, endpoint.stop_code)
                log.info(msg)
                if endpoint.stop_code > 0:
                    self._abort(endpoint.stop_code, msg)
            if self._control_transfer is not None:
                self._control_transfer.process()
                if self._control_transfer.stop_code is not None and self._control_transfer.stop_code > 0:
                    msg = 'Control pipe %r stopped: %s' % (pipe_id, self._control_transfer.stop_code)
                    log.info(msg)
                    self._abort(self._control_transfer.stop_code, msg)


def scan(name: str) -> List[WinUsbDevice]:
    """Scan for attached devices.

    :param name: The case-insensitive name of the device to scan.
    :return: The list of discovered :class:`WinUsbDevice` instances.
    """
    paths = []
    infos = INFO[name.lower()]
    for info in infos:
        guid = info['DeviceInterfaceGUID']
        p = device_interface_guid_to_paths(guid)
        paths.extend(p)
    if not len(paths):
        log.debug('scan found no devices')
        return []
    log.debug('scan found %s' % paths)
    devices = [WinUsbDevice(p) for p in paths]
    return devices



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the JouleScope device

This test requires that JouleScope hardware be attached to this PC!
"""

import unittest
from joulescope.driver import scan, UsbdRequest, LOOPBACK_BUFFER_SIZE, Device
from joulescope.usb import hw_tests
from joulescope.pattern_buffer import PatternBuffer


class TestPattern(unittest.TestCase):

    def setUp(self):
        self.device = None
        self.devices = scan(name='joulescope')
        if not len(self.devices):
            raise unittest.SkipTest("no devices found")
        if len(self.devices) > 1:
            print("multiple devices found")
        self.device = self.devices[0]
        self.device.open()

    def tearDown(self):
        if self.device is not None:
            self.device.close()

    def test_control_loopback_wvalue(self):
        usb = self.device.usb_device
        hw_tests.control_loopback_wvalue(usb, UsbdRequest.LOOPBACK_WVALUE, 17)

    def test_control_loopback_buffer(self):
        usb = self.device.usb_device
        hw_tests.control_loopback_buffer(usb, UsbdRequest.LOOPBACK_BUFFER, LOOPBACK_BUFFER_SIZE, 4)

    def _pattern(self, duration=None):
        duration = int(duration) if duration is not None else 1.0
        buffer = PatternBuffer()
        self.device.stream_buffer = buffer
        self.device.read(duration=duration, out_format='raw')
        s = buffer.status()
        self.assertGreater(s['sample_id'], 1000000)
        self.assertEqual(s['header_error'], 0)
        self.assertLessEqual(s['pkt_index_error'], 1)
        self.assertLessEqual(s['pattern_error'], 1)

    def test_datapath_usb(self):
        self.device.parameter_set('control_test_mode', 'usb')
        self.device.parameter_set('source', 'pattern_usb')
        self._pattern(1.0)

    def test_datapath_sensor(self):
        self.device.parameter_set('control_test_mode', 'normal')
        self.device.parameter_set('source', 'pattern_sensor')
        self._pattern(1.0)

    def test_read(self):
        self.device.parameter_set('source', 'raw')
        for i in range(10):
            v = self.device.read(contiguous_duration=0.125, out_format='raw')
            self.assertEqual((self.device.sampling_frequency // 8, 2), v.shape)


class TestAttributes(unittest.TestCase):

    def test_defaults(self):
        d = Device(None)
        self.assertEqual(2000000, d.sampling_frequency)
        self.assertEqual('30 seconds', d.parameter_get('buffer_duration'))
        self.assertEqual(30, d.parameter_get('buffer_duration', dtype='actual'))
        self.assertEqual('2 Hz', d.parameter_get('reduction_frequency'))
        self.assertIsNone(d.stream_buffer)
        self.assertIsNone(d.calibration)
        self.assertEqual('i_range', d.parameters(name='i_range').name)
        self.assertEqual('off', d.parameter_get('i_range'))
        with self.assertRaises(Exception):
            d.serial_number
        with self.assertRaises(Exception):
            d.info()
        self.assertFalse(d.is_streaming)
        with self.assertRaises(Exception):
            d.status()
        with self.assertRaises(Exception):
            d.extio_status()

    def test_parameter_arbitrary(self):
        d = Device(None)
        d.parameter_set('buffer_duration', 2)

    def test_reduction_frequency(self):
        d = Device(None)
        for frequency in [1, 2, 4, 10, 20, 50, 100]:
            with self.subTest(frequency=frequency):
                d.reduction_frequency = frequency
                self.assertEqual(frequency, d.reduction_frequency)

    def test_info(self):
        d = Device(None)
        self.assertEqual('JS110', d.parameter_get('model'))
        self.assertEqual(None, d.parameter_get('device_serial_number'))
        self.assertEqual(None, d.parameter_get('hardware_serial_number'))



# Copyright 2020 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import numpy as np
from joulescope.stream_buffer import RunningStatistics as Statistics
from joulescope.stream_buffer import stats_compute, stats_array_factory



class TestStatistics(unittest.TestCase):

    def test_initialize_empty(self):
        s = Statistics()
        self.assertEqual(0, len(s))
        self.assertEqual(0.0, s.mean)
        self.assertEqual(0.0, s.var)
        self.assertEqual(0.0, s.std)

    def test_add_zero_once(self):
        s = Statistics()
        s.add(0.0)
        self.assertEqual(1, len(s))
        self.assertEqual(0.0, s.mean)
        self.assertEqual(0.0, s.var)
        self.assertEqual(0.0, s.std)
        self.assertEqual(0.0, s.min)
        self.assertEqual(0.0, s.max)

    def test_add_zero_twice(self):
        s = Statistics()
        s.add(0.0)
        s.add(0.0)
        self.assertEqual(2, len(s))
        self.assertEqual(0.0, s.mean)
        self.assertEqual(0.0, s.var)
        self.assertEqual(0.0, s.std)
        self.assertEqual(0.0, s.min)
        self.assertEqual(0.0, s.max)

    def test_add_multiple(self):
        s = Statistics()
        s.add(0.0)
        s.add(1.0)
        s.add(2.0)
        self.assertEqual(3, len(s))
        self.assertEqual(1.0, s.mean)
        # note that np.var returns the variance, not sample variance
        sample_variance = 3.0 / 2.0 * np.var([0.0, 1.0, 2.0])
        self.assertEqual(1.0, sample_variance)
        self.assertEqual(sample_variance, s.var)
        self.assertEqual(np.sqrt(sample_variance), s.std)
        self.assertEqual(0.0, s.min)
        self.assertEqual(2.0, s.max)

    def test_combine_both_empty(self):
        s1 = Statistics()
        s2 = Statistics()
        s3 = s1.combine(s2)
        self.assertEqual(0, len(s3))

    def test_combine_self_empty(self):
        s1 = Statistics()
        s2 = Statistics()
        s2.add(1.0)
        s3 = s1.combine(s2)
        self.assertEqual(1, len(s3))
        self.assertEqual(1.0, s3.mean)

    def test_combine_other_empty(self):
        s1 = Statistics()
        s2 = Statistics()
        s1.add(1.0)
        s3 = s1.combine(s2)
        self.assertEqual(1, len(s3))
        self.assertEqual(1.0, s3.mean)

    def test_combine(self):
        s1 = Statistics()
        for x in np.arange(0, 11):
            s1.add(x)
        s2 = Statistics()
        for x in np.arange(11, 21):
            s2.add(x)
        s3 = Statistics()
        for x in np.arange(0, 21):
            s3.add(x)
        sc = s1.combine(s2)
        self.assertEqual(len(s3), len(sc))
        self.assertEqual(s3.mean, sc.mean)
        self.assertEqual(s3.var, sc.var)
        self.assertEqual(s3.std, sc.std)
        self.assertEqual(s3.min, sc.min)
        self.assertEqual(s3.max, sc.max)

    def test_stats_compute(self):
        v = np.arange(0, 11, dtype=np.float32)
        s_array = stats_array_factory(1)
        stats_compute(v, s_array[0, 0:1])
        self.assertEqual(np.mean(v), s_array[0, 0]['mean'])



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time

log = logging.getLogger(__name__)


class NotificationHandler:

    def __init__(self, period=None):
        self.period = period
        self._time_last = 0.0
        self._notify = []

    def emit(self, obj=None):
        if self.period is not None:
            time_now = time.time()
            time_delta = time_now - self._time_last
            if time_delta < self.period:
                return
            self._time_last = time_now
        self.emit_always(obj)

    def emit_always(self, obj=None):
        for fn in self._notify:
            try:
                fn(obj)
            except Exception:
                log.exception('while notify %s' % fn.__name__)

    def register(self, fn):
        """Attach a function to be notified of changes

        :param fn: The callback function which is called when the
            data buffer content changes.
        """
        if fn not in self._notify:
            self._notify.append(fn)
        else:
            log.info('NotificationHandler.register already has %s' % fn.__name__)

    def unregister(self, fn):
        if fn in self._notify:
            self._notify.remove(fn)
        else:
            log.info('NotificationHandler.unregister does not contain %s' % fn.__name__)



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import datafile
from joulescope.calibration import Calibration
from joulescope.stream_buffer import reduction_downsample, Statistics, stats_to_api, \
    stats_invalidate, \
    stats_factory, stats_array_factory, stats_array_clear, stats_array_invalidate, \
    STATS_DTYPE, STATS_FIELD_NAMES, STATS_FIELD_COUNT, NP_STATS_NAMES, \
    I_RANGE_MISSING, SUPPRESS_SAMPLES_MAX, RawProcessor, stats_compute
from joulescope import array_storage
import json
import numpy as np
import datetime
import logging

log = logging.getLogger(__name__)

DATA_RECORDER_FORMAT_VERSION = '2'
SAMPLES_PER_REDUCTION = 20000   # 100 Hz @ 2 MSPS
REDUCTIONS_PER_TLV = 20         # 5 Hz @ 2 MSPS
TLVS_PER_BLOCK = 5              # 1 Hz @ 2 MSPS

_STATS_VALUES_V1 = 4


_SIGNALS_UNITS = {
    'current': 'A',
    'voltage': 'V',
    'power': 'W',
    'raw_current': 'LSBs',
    'raw_voltage': 'LSBs',
    'raw': '',
    'bits': '',
    'current_range': '',
    'current_lsb': '',
    'voltage_lsb': '',
}


_DOWNSAMPLE_FORMATTERS = {
    0: lambda x: x.astype(dtype=np.float32),
    1: lambda x: x.astype(dtype=np.float32),
    2: lambda x: x.astype(dtype=np.float32),
    3: lambda x: (x * 16).astype(dtype=np.uint8),
    4: lambda x: (x * 15).astype(dtype=np.uint8),
    5: lambda x: (x * 15).astype(dtype=np.uint8),
}


_DOWNSAMPLE_UNFORMATTERS = {
    0: lambda x: x,
    1: lambda x: x,
    2: lambda x: x,
    3: lambda x: x.astype(dtype=np.float32) * (1.0 / 16),
    4: lambda x: x.astype(dtype=np.float32) * (1.0 / 15),
    5: lambda x: x.astype(dtype=np.float32) * (1.0 / 15),
}


def construct_record_filename():
    time_start = datetime.datetime.utcnow()
    timestamp_str = time_start.strftime('%Y%m%d_%H%M%S')
    return f'{timestamp_str}.jls'


class DataRecorder:
    """Record Joulescope data to a file.

    :param filehandle: The file-like object or file name.
    :param calibration: The calibration bytes in datafile format.
        None (default) uses the unit gain calibration.
    :param user_data: Arbitrary JSON-serializable user data that is
        added to the file.
    """
    def __init__(self, filehandle, calibration=None, user_data=None):

        log.info('init')
        if isinstance(filehandle, str):
            self._fh = open(filehandle, 'wb')
            filehandle = self._fh
        else:
            self._fh = None

        self._sampling_frequency = 0
        self._samples_per_tlv = 0
        self._samples_per_block = 0
        self._samples_per_reduction = 0
        self._reductions_per_tlv = 0
        self._reduction = None

        self._sample_id_tlv = None  # sample id for start of next TLV
        self._sample_id_block = None  # sample id for start of current block, None if not started yet

        self._stream_buffer = None  # to ensure same
        self._sb_sample_id_last = None
        self._voltage_range = None
        self._data_buffer = []
        self._sample_buffers = {}

        self._writer = datafile.DataFileWriter(filehandle)
        self._closed = False
        self._total_size = 0

        if user_data is not None:
            b = json.dumps(user_data).encode('utf-8')
            self._writer.append(datafile.TAG_USER_JSON, b)
        if calibration is not None:
            if isinstance(calibration, Calibration):
                calibration = calibration.data
            self._writer.append_subfile('calibration', calibration)

    def _initialize(self, sampling_frequency, data_format):
        self._sampling_frequency = sampling_frequency
        self._samples_per_reduction = SAMPLES_PER_REDUCTION
        self._samples_per_tlv = self._samples_per_reduction * REDUCTIONS_PER_TLV
        self._samples_per_block = self._samples_per_tlv * TLVS_PER_BLOCK

        # dependent vars
        self._reductions_per_tlv = self._samples_per_tlv // self._samples_per_reduction
        reduction_block_size = self._samples_per_block // self._samples_per_reduction

        self._reduction = stats_array_factory(reduction_block_size)
        self._append_configuration(data_format)
        self._writer.collection_start(0, 0)

    def _append_configuration(self, data_format=None):
        data_format = 'none' if data_format is None else str(data_format)
        config = {
            'type': 'config',
            'data_recorder_format_version': DATA_RECORDER_FORMAT_VERSION,
            'sampling_frequency': self._sampling_frequency,
            'samples_per_reduction': self._samples_per_reduction,
            'samples_per_tlv': self._samples_per_tlv,
            'samples_per_block': self._samples_per_block,
            'reduction_fields': ['current', 'voltage', 'power',
                                 'current_range', 'current_lsb', 'voltage_lsb'],
            'data_format': data_format,
            'reduction_format': 'v2',
        }
        cfg_data = json.dumps(config).encode('utf-8')
        self._writer.append(datafile.TAG_META_JSON, cfg_data)

    def _collection_start(self, data=None):
        log.debug('_collection_start()')
        c = self._writer.collection_start(1, 0, data=data)
        c.metadata = {'start_sample_id': self._sample_id_tlv}
        c.on_end = self._collection_end
        self._sample_id_block = self._sample_id_tlv

    def _collection_end(self, collection):
        sample_count = (self._sample_id_tlv - self._sample_id_block)
        r_stop = sample_count // self._samples_per_reduction
        log.debug('_collection_end(%s, %s)', r_stop, len(self._reduction))
        data = self._reduction[:r_stop, :]
        arrays = {0x00: data[:, 0]['length'].astype(np.uint64)}
        for field_idx, field in enumerate(STATS_FIELD_NAMES):
            for stat_idx, stat in enumerate(NP_STATS_NAMES[1:]):
                x_id = ((field_idx & 0x0f) << 4) | ((stat_idx + 1) & 0x0f)
                x = data[:, field_idx][stat]
                arrays[x_id] = x.astype(np.float32)
        value = array_storage.pack(arrays, r_stop)
        value = value.tobytes()
        self._writer.collection_end(collection, value)
        self._sample_id_block = None
        stats_array_clear(self._reduction)

    def stream_notify(self, stream_buffer):
        """Process data from a stream buffer.

        :param stream_buffer: The stream_buffer instance which has:
            * "output_sampling_frequency" member -> float
            * "has_raw" member -> in [True, False]
            * "sample_id_range" member => (start, stop)
            * "voltage_range" member -> in [0, 1]
            * samples_get(start_sample_id, stop_sample_id, dtype)
            * data_get(start_sample_id, stop_sample_id, increment, out)
            * __len__ : maximum number of samples that stream_buffer holds
        """
        finalize = False
        if stream_buffer is None:
            if self._stream_buffer is None:
                return
            finalize = True
            stream_buffer = self._stream_buffer
        sb_start, sb_stop = stream_buffer.sample_id_range
        if self._stream_buffer is None:
            self._stream_buffer = stream_buffer
            data_format = 'raw' if stream_buffer.has_raw else 'float32_v2'
            self._initialize(stream_buffer.output_sampling_frequency, data_format)
            self._sample_id_tlv = sb_stop
            self._sample_id_block = None
            if self._samples_per_tlv > len(stream_buffer):
                raise ValueError('stream_buffer length too small.  %s > %s' %
                                 (self._samples_per_tlv, len(stream_buffer)))
        elif self._stream_buffer != stream_buffer:
            raise ValueError('Supports only a single stream_buffer instance')
        if sb_start > self._sample_id_tlv:
            raise ValueError('stream_buffer does not contain sample_id')

        while True:
            finalize_now = False
            sample_id_next = self._sample_id_tlv + self._samples_per_tlv
            if sb_stop < sample_id_next:
                if finalize and self._sample_id_tlv < sb_stop:
                    finalize_now = True
                    sample_id_next = sb_stop
                else:
                    break

            self._voltage_range = stream_buffer.voltage_range
            if self._sample_id_block is None:
                collection_data = {
                    'v_range': self._voltage_range,
                    'sample_id': self._sample_id_tlv,
                }
                collection_data = json.dumps(collection_data).encode('utf-8')
                self._collection_start(data=collection_data)

            log.debug('_process() add tlv %d', self._sample_id_tlv)
            if stream_buffer.has_raw:
                self._append_raw_data(self._sample_id_tlv, sample_id_next)
            else:
                self._append_float_data(self._sample_id_tlv, sample_id_next)
            self._total_size += (sample_id_next - self._sample_id_tlv)
            tlv_offset = (self._sample_id_tlv - self._sample_id_block) // self._samples_per_tlv
            r_start = tlv_offset * self._reductions_per_tlv
            r_stop = r_start + self._reductions_per_tlv
            stream_buffer.data_get(self._sample_id_tlv, sample_id_next,
                                   self._samples_per_reduction, out=self._reduction[r_start:r_stop, :])
            self._sample_id_tlv = sample_id_next
            if finalize_now or self._sample_id_tlv - self._sample_id_block >= self._samples_per_block:
                self._collection_end(self._writer.collections[-1])

    def _append_raw_data(self, start, stop):
        if self._closed:
            return
        b = self._stream_buffer.samples_get(start, stop, fields='raw')
        data = b.tobytes()
        self._writer.append(datafile.TAG_DATA_BINARY, data, compress=False)

    def _append_float_data(self, start, stop):
        if self._closed:
            return
        data = self._stream_buffer.samples_get(start, stop, fields=STATS_FIELD_NAMES)
        arrays = {}
        for field_idx, field in enumerate(STATS_FIELD_NAMES):
            x_id = (field_idx << 4) | 0x01  # stat_idx=mean
            x = data['signals'][field]['value']
            x = _DOWNSAMPLE_FORMATTERS[field_idx](x)
            arrays[x_id] = x
        data = array_storage.pack(arrays, stop - start)
        data = data.tobytes()
        self._writer.append(datafile.TAG_DATA_BINARY, data, compress=False)

    def _insert_remove_processed_data(self):
        while len(self._data_buffer) and self._data_buffer[0]['time']['sample_id_range']['value'][1] < self._sample_id_tlv:
            self._data_buffer.pop()

    def _insert_size_pending(self):
        self._insert_remove_processed_data()
        if not len(self._data_buffer):
            return 0
        sz = np.sum([x['time']['samples']['value'] for x in self._data_buffer])
        start_idx = self._data_buffer[0]['time']['sample_id_range']['value'][0]
        sz -= self._sample_id_tlv - start_idx
        return sz

    def _insert_next(self):
        # Will write any pending data up to a full data TLV
        # WARNING: does not check for sufficient data to fill a data TLV
        # since it needs to also handle the final partial.
        if self._sample_id_block is None:
            if not len(self._data_buffer):
                return
            has_raw = 'raw' in self._data_buffer[0]['signals']
            collection_data = {'sample_id': self._sample_id_tlv}
            if has_raw:
                collection_data['v_range'] = self._data_buffer[0]['signals']['raw']['voltage_range']
            collection_data = json.dumps(collection_data).encode('utf-8')
            self._collection_start(data=collection_data)
        offset = 0
        finalize_now = True
        while len(self._data_buffer):
            do_pop = True
            data = self._data_buffer[0]
            s0, s1 = data['time']['sample_id_range']['value']  # samples, not indices
            assert(self._sample_id_tlv + offset >= s0)
            idx0 = self._sample_id_tlv + offset - s0  # input buffer starting index
            idx1 = s1 - s0  # input buffer ending index
            if s1 > self._sample_id_tlv + self._samples_per_tlv:
                idx1 = self._sample_id_tlv + self._samples_per_tlv - s0
                do_pop = False
            idx_end = offset + idx1 - idx0
            for field, d in self._sample_buffers.items():
                if field == 'raw':
                    self._sample_buffers[field][offset:idx_end, :] = data['signals'][field]['value'][idx0:idx1, :]
                else:
                    self._sample_buffers[field][offset:idx_end] = data['signals'][field]['value'][idx0:idx1]
            offset += idx1 - idx0
            if do_pop:
                self._data_buffer.pop(0)
            if offset >= self._samples_per_tlv:
                finalize_now = False
                break

        # write the data TLV
        if 'raw' in self._sample_buffers:
            data = self._sample_buffers['raw'][:offset, :].tobytes()
            self._writer.append(datafile.TAG_DATA_BINARY, data, compress=False)
        else:
            arrays = {}
            for field_idx, field in enumerate(STATS_FIELD_NAMES):
                x_id = (field_idx << 4) | 0x01  # stat_idx=mean
                x = self._sample_buffers[field][:offset]
                x = _DOWNSAMPLE_FORMATTERS[field_idx](x)
                arrays[x_id] = x
            data = array_storage.pack(arrays, self._samples_per_tlv)
            data = data.tobytes()
            self._writer.append(datafile.TAG_DATA_BINARY, data, compress=False)

        # update reductions
        tlv_offset = (self._sample_id_tlv - self._sample_id_block) // self._samples_per_tlv
        r_start = tlv_offset * self._reductions_per_tlv
        r_stop = r_start + (offset * self._reductions_per_tlv) // self._samples_per_tlv
        for field_idx, field in enumerate(STATS_FIELD_NAMES):
            d = self._sample_buffers[field]
            for idx in range(0, r_stop - r_start):
                r = r_start + idx
                k0 = idx * self._samples_per_reduction
                k1 = k0 + self._samples_per_reduction
                stats_compute(d[k0:k1], self._reduction[r, field_idx:field_idx + 1])

        self._sample_id_tlv += self._samples_per_tlv
        self._total_size += offset
        if finalize_now or self._sample_id_tlv - self._sample_id_block >= self._samples_per_block:
            self._collection_end(self._writer.collections[-1])
        return

    def insert(self, data):
        """Insert sample data.

        :param data: A dict in the :meth:`StreamBuffer.samples_get` format.
        """
        if self._sample_id_tlv is None:  # first call
            data_format = 'raw' if 'raw' in data['signals'] else 'float32_v2'
            self._initialize(data['time']['sampling_frequency']['value'], data_format)
            self._sample_id_tlv = data['time']['sample_id_range']['value'][0]
            self._sample_buffers = {}
            for field in data['signals'].keys():
                if field == 'raw':
                    d = np.empty((self._samples_per_tlv, 2), dtype=np.uint16)
                else:
                    d = np.empty(self._samples_per_tlv, dtype=np.float32)
                self._sample_buffers[field] = d
        if data is not None:
            self._data_buffer.append(data)
        while True:
            if self._insert_size_pending() < self._samples_per_tlv:
                break
            self._insert_next()
        if data is None:
            self._insert_next()  # short data finalize

    def _append_meta(self, footer_user_data=None):
        index = {
            'type': 'footer',
            'size': self._total_size,  # in samples
        }
        data = json.dumps(index).encode('utf-8')
        self._writer.append(datafile.TAG_META_JSON, data)
        if footer_user_data is not None:
            data = json.dumps(footer_user_data).encode('utf-8')
            self._writer.append(datafile.TAG_USER_JSON, data)

    def close(self, footer_user_data=None):
        """Finalize and close the recording."""
        if self._closed:
            return
        if self._sample_id_tlv is None:
            self._append_configuration()
        if self._stream_buffer:
            self.stream_notify(None)
        if self._data_buffer:
            self.insert(None)
        self._closed = True
        while len(self._writer.collections):
            collection = self._writer.collections[-1]
            if len(collection.metadata):
                self._collection_end(collection)
            else:
                self._writer.collection_end()
        self._append_meta(footer_user_data)
        self._writer.finalize()
        if self._fh is not None:
            self._fh.close()
            self._fh = None


class DataReader:
    """Read Joulescope data from a file."""

    def __init__(self):
        self.calibration = None
        self.config = None
        self.footer = None
        self._fh_close = False
        self._fh = None
        self._f = None  # type: datafile.DataFileReader
        self._data_start_position = 0
        self._voltage_range = 0
        self._sample_cache = None
        self._data_get_handler = None
        self._reduction_handler = None
        self._samples_get_handler = None
        self._statistics_get_handler = None
        self.raw_processor = RawProcessor()
        self.user_data = None
        self.footer_user_data = None

    def __str__(self):
        if self._f is not None:
            return 'DataReader %.2f seconds (%d samples)' % (self.duration, self.footer['size'])

    def close(self):
        """Close the recording file."""
        if self._fh_close:
            self._fh.close()
        self._fh_close = False
        self._fh = None
        self._f = None
        self._sample_cache = None
        self._reduction_cache = None

    def open(self, filehandle):
        """Open a recording file.

        :param filehandle: The seekable filehandle or filename string.
        """
        self.close()
        self.calibration = Calibration()  # default calibration
        self.config = None
        self.footer = None
        self._data_start_position = 0

        if isinstance(filehandle, str):
            log.info('DataReader(%s)', filehandle)
            self._fh = open(filehandle, 'rb')
            self._fh_close = True
        else:
            self._fh = filehandle
            self._fh_close = False
        self._f = datafile.DataFileReader(self._fh)
        while True:
            tag, value = self._f.peek()
            if tag is None:
                if self._data_start_position and self.config is not None:
                    log.warning('Unexpected file truncation, attempting recovery')
                    sample_count = self._sample_count()
                    log.warning('Recovery found %d samples', sample_count)
                    self.footer = {
                        'type': 'footer',
                        'size': sample_count,  # in samples
                    }
                else:
                    raise ValueError('could not read file')
                break
            elif tag == datafile.TAG_SUBFILE:
                name, data = datafile.subfile_split(value)
                if name == 'calibration':
                    self.calibration = Calibration().load(data)
            elif tag == datafile.TAG_COLLECTION_START:
                self._data_start_position = self._f.tell()
            elif tag == datafile.TAG_META_JSON:
                meta = json.loads(value.decode('utf-8'))
                type_ = meta.get('type')
                if type_ == 'config':
                    self._configure(meta)
                elif type_ == 'footer':
                    self.footer = meta
                else:
                    log.warning('Unknown JSON section type=%s', type_)
            elif tag == datafile.TAG_USER_JSON:
                if self.footer is None:
                    self.user_data = json.loads(value.decode('utf-8'))
                else:
                    self.footer_user_data = json.loads(value.decode('utf-8'))
            elif tag == datafile.TAG_END:
                break
            self._f.skip()
        if self.config is None or self.footer is None:
            raise ValueError('could not read file')
        log.info('DataReader with %d samples:\n%s', self.footer['size'], json.dumps(self.config, indent=2))
        if self._data_start_position == 0 and self.footer['size']:
            raise ValueError(f"data not found, but expect {self.footer['size']} samples")
        if int(self.config['data_recorder_format_version']) > int(DATA_RECORDER_FORMAT_VERSION):
            raise ValueError('Invalid file format')
        self.config.setdefault('reduction_fields', ['current', 'voltage', 'power'])
        cal = self.calibration
        self.raw_processor.calibration_set(cal.current_offset, cal.current_gain, cal.voltage_offset, cal.voltage_gain)
        return self

    def _configure(self, config):
        config.setdefault('data_format', 'raw')
        config.setdefault('reduction_format', 'v1')
        self.config = config
        self._samples_get_handler = getattr(self, '_samples_get_handler_' + config['data_format'])
        self._data_get_handler = getattr(self, '_data_get_handler_' + config['data_format'])
        self._statistics_get_handler = getattr(self, '_statistics_get_handler_' + config['data_format'])
        self._reduction_handler = getattr(self, '_reduction_handler_' + config['reduction_format'])

    @property
    def sample_id_range(self):
        """The sample ID range.

        :return: [start, stop] sample identifiers.
        """
        if self._f is not None:
            s_start = 0
            s_end = int(s_start + self.footer['size'])
            return [s_start, s_end]
        return [0, 0]

    @property
    def sampling_frequency(self):
        """The output sampling frequency."""
        f = 0.0
        if self._f is not None:
            f = float(self.config['sampling_frequency'])
        if f <= 0.0:
            log.warning('Invalid sampling frequency %r, assume 1.0 Hz.', f)
            f = 1.0
        return f

    @property
    def input_sampling_frequency(self):
        """The original input sampling frequency."""
        f = 0.0
        if self._f is None:
            pass
        elif 'input_sampling_frequency' in self.config:
            f = float(self.config['sampling_frequency'])
        else:
            f = self.sampling_frequency
        if f <= 0.0:
            log.warning('Invalid input sampling frequency %r, assume 1.0 Hz.', f)
            f = 1.0
        return f

    @property
    def output_sampling_frequency(self):
        """The output sampling frequency."""
        return self.sampling_frequency

    @property
    def reduction_frequency(self):
        """The reduction frequency or 1 if no reduction."""
        f = 0.0
        try:
            if self._f is not None:
                f = self.config['sampling_frequency'] / self.config['samples_per_reduction']
        except Exception:
            log.warning('Could not get reduction frequency.')
        if f <= 0.0:
            log.warning('Invalid input sampling frequency %r, assume 1.0 Hz.', f)
            f = 1.0
        return f

    @property
    def duration(self):
        """The data file duration, in seconds."""
        f = self.sampling_frequency
        if f > 0:
            r = self.sample_id_range
            return (r[1] - r[0]) / f
        return 0.0

    @property
    def voltage_range(self):
        """The data file voltage range."""
        return self._voltage_range

    def _sample_count(self):
        """Count the actual samples in the file.

        WARNING: this operation may be slow.  Use footer when possible.
        """
        samples_per_block = self.config['samples_per_block']
        sample_count = 0
        self._fh.seek(self._data_start_position)

        while True:
            tag, _ = self._f.peek_tag_length()
            if tag is None:
                break
            if tag == datafile.TAG_COLLECTION_START:
                self._f.skip()
                sample_count += samples_per_block
            else:
                self._f.advance()
        return sample_count

    def _validate_range(self, start=None, stop=None, increment=None):
        idx_start, idx_end = self.sample_id_range
        if increment is not None:
            idx_end = ((idx_end + increment - 1) // increment) * increment
        # log.debug('[%d, %d] : [%d, %d]', start, stop, idx_start, idx_end)
        if start == idx_start and start == stop:
            pass  # empty is allowed
        elif not idx_start <= start < idx_end:
            raise ValueError('start out of range: %d <= %d < %d' % (idx_start, start, idx_end))
        elif not idx_start <= stop <= idx_end:
            raise ValueError('stop out of range: %d <= %d <= %d: %s' %
                             (idx_start, stop, idx_end, increment))
        if stop < start:
            stop = start
        return start, stop

    def _sample_tlv(self, sample_idx):
        """Get the sample data entry for the TLV containing sample_idx.

        :param sample_idx: The sample index.
        :return: The dict containing the data.
        """

        if self._sample_cache and self._sample_cache['start'] <= sample_idx < self._sample_cache['stop']:
            # cache hit
            return self._sample_cache

        idx_start, idx_end = self.sample_id_range
        if not idx_start <= sample_idx < idx_end:
            raise ValueError('sample index out of range: %d <= %d < %d' % (idx_start, sample_idx, idx_end))

        if self._sample_cache is not None:
            log.debug('_sample_cache cache miss: %s : %s %s',
                      sample_idx, self._sample_cache['start'], self._sample_cache['stop'])

        # seek
        samples_per_tlv = self.config['samples_per_tlv']
        samples_per_block = self.config['samples_per_block']
        tgt_block = sample_idx // samples_per_block

        if self._sample_cache is not None and sample_idx > self._sample_cache['start']:
            # continue forward
            self._fh.seek(self._sample_cache['tlv_pos'])
            voltage_range = self._sample_cache['voltage_range']
            block_fh_pos = self._sample_cache['block_pos']
            current_sample_idx = self._sample_cache['start']
            block_counter = current_sample_idx // samples_per_block
        else:  # add case for rewind?
            log.debug('_sample_tlv resync to beginning')
            self._fh.seek(self._data_start_position)
            voltage_range = 0
            block_fh_pos = 0
            block_counter = 0
            current_sample_idx = 0
            if self._f.advance() != datafile.TAG_COLLECTION_START:
                raise ValueError('data section must be single collection')

        while True:
            tag, _ = self._f.peek_tag_length()
            if tag is None:
                log.error('sample_tlv not found before end of file: %s > %s', sample_idx, current_sample_idx)
                break
            if tag == datafile.TAG_COLLECTION_START:
                if block_counter < tgt_block:
                    self._f.skip()
                    block_counter += 1
                else:
                    block_fh_pos = self._f.tell()
                    tag, collection_bytes = next(self._f)
                    c = datafile.Collection.decode(collection_bytes)
                    if c.data:
                        collection_start_meta = json.loads(c.data)
                        voltage_range = collection_start_meta.get('v_range', 0)
                        self._voltage_range = voltage_range
                    current_sample_idx = block_counter * samples_per_block
            elif tag == datafile.TAG_COLLECTION_END:
                block_counter += 1
                self._f.advance()
            elif tag == datafile.TAG_DATA_BINARY:
                tlv_stop = current_sample_idx + samples_per_tlv
                if current_sample_idx <= sample_idx < tlv_stop:
                    # found it!
                    tlv_pos = self._f.tell()
                    tag, value = next(self._f)
                    self._sample_cache = {
                        'voltage_range': voltage_range,
                        'start': current_sample_idx,
                        'stop': tlv_stop,
                        'value': value,
                        'tlv_pos': tlv_pos,
                        'block_pos': block_fh_pos,
                    }
                    return self._sample_cache
                else:
                    self._f.advance()
                current_sample_idx = tlv_stop
            else:
                self._f.advance()

    def _raw(self, start=None, stop=None):
        """Get the raw data.

        :param start: The starting sample (must already be validated).
        :param stop: The ending sample (must already be validated).
        :return: The output which is (out_raw, bits, out_cal).
        """
        x_start, x_stop = self.sample_id_range
        self._fh.seek(self._data_start_position)
        self._validate_range(start, stop)
        length = stop - start
        if length <= 0:
            return np.empty((0, 2), dtype=np.uint16)
        # process extra before & after to handle filtering
        if start > SUPPRESS_SAMPLES_MAX:
            sample_idx = start - SUPPRESS_SAMPLES_MAX
            prefix_count = SUPPRESS_SAMPLES_MAX
        else:
            sample_idx = 0
            prefix_count = start
        if stop + SUPPRESS_SAMPLES_MAX <= x_stop:
            end_idx = stop + SUPPRESS_SAMPLES_MAX
        else:
            end_idx = x_stop
        out_idx = 0

        d_raw = np.empty((end_idx - sample_idx, 2), dtype=np.uint16)
        if self._f.advance() != datafile.TAG_COLLECTION_START:
            raise ValueError('data section must be single collection')
        while sample_idx < end_idx:
            sample_cache = self._sample_tlv(sample_idx)
            if sample_cache is None:
                break
            value = sample_cache['value']
            data = np.frombuffer(value, dtype=np.uint16).reshape((-1, 2))
            b_start = sample_idx - sample_cache['start']
            length = sample_cache['stop'] - sample_cache['start'] - b_start
            out_remaining = end_idx - sample_idx
            length = min(length, out_remaining)
            if length <= 0:
                break
            b_stop = b_start + length
            d = data[b_start:b_stop, :]
            d_raw[out_idx:(out_idx + length), :] = d
            out_idx += length
            sample_idx += length

        d_raw = d_raw[:out_idx, :]
        self.raw_processor.reset()
        self.raw_processor.voltage_range = self._voltage_range
        d_cal, d_bits = self.raw_processor.process_bulk(d_raw.reshape((-1, )))
        j = prefix_count
        k = min(prefix_count + stop - start, out_idx)
        return d_raw[j:k, :], d_bits[j:k], d_cal[j:k, :]

    def _downsampled(self, start, stop, fields):
        """Get the _downsampled data.

        :param start: The starting sample (must already be validated).
        :param stop: The ending sample (must already be validated).
        :return: The output which is dict of field name to values.
        """
        self._fh.seek(self._data_start_position)
        start, stop = self._validate_range(start, stop)
        length = stop - start
        field_idxs = []
        if fields is None:
            fields = ['current', 'voltage', 'power', 'current_range', 'current_lsb', 'voltage_lsb']
        for field in fields:
            if field not in STATS_FIELD_NAMES:
                raise ValueError(f'Field {field} not available')
            idx = (STATS_FIELD_NAMES.index(field) << 4) | 1
            field_idxs.append((field, idx))
        rv = {}
        for _, field_idx in field_idxs:
            rv[field_idx] = np.empty(length, dtype=np.float32)
        out_idx = 0
        while start < stop:
            sample_cache = self._sample_tlv(start)
            if sample_cache is None:
                break
            v = np.frombuffer(sample_cache['value'], dtype=np.uint8)
            value, sample_count = array_storage.unpack(v)
            b_start = start - sample_cache['start']
            length = sample_cache['stop'] - sample_cache['start'] - b_start
            out_remaining = stop - start
            length = min(length, out_remaining)
            if length <= 0:
                break
            b_stop = b_start + length
            for _, field_idx in field_idxs:
                rv[field_idx][out_idx:(out_idx + length)] = value[field_idx][b_start:b_stop]
            out_idx += length
            start += length

        result = {}
        for field, field_idx in field_idxs:
            result[field] = rv[field_idx][:out_idx]
        return result

    def _reduction_tlv(self, reduction_idx):
        sz = self.config['samples_per_reduction']
        incr = self.config['samples_per_block'] // sz
        tgt_r_idx = reduction_idx

        if self._reduction_cache and self._reduction_cache['r_start'] <= tgt_r_idx < self._reduction_cache['r_stop']:
            return self._reduction_cache

        if self._reduction_cache is not None:
            log.debug('_reduction_tlv cache miss: %s : %s %s',
                      tgt_r_idx, self._reduction_cache['r_start'], self._reduction_cache['r_stop'])

        idx_start, idx_end = self.sample_id_range
        r_start = idx_start // sz
        r_stop = idx_end // sz
        if not r_start <= tgt_r_idx < r_stop:
            raise ValueError('reduction index out of range: %d <= %d < %d', r_start, tgt_r_idx, r_stop)

        if self._reduction_cache is not None and tgt_r_idx > self._reduction_cache['r_start']:
            # continue forward
            self._fh.seek(self._reduction_cache['next_collection_pos'])
            r_idx = self._reduction_cache['r_stop']
        else:  # add case for rewind?
            log.debug('_reduction_tlv resync to beginning')
            self._fh.seek(self._data_start_position)
            r_idx = 0
            if self._f.advance() != datafile.TAG_COLLECTION_START:
                raise ValueError('data section must be single collection')
            self._fh.seek(self._data_start_position)
            if self._f.advance() != datafile.TAG_COLLECTION_START:
                raise ValueError('data section must be single collection')

        while True:
            tag, _ = self._f.peek_tag_length()
            if tag is None or tag == datafile.TAG_COLLECTION_END:
                log.error('reduction_tlv not found before end of file: %s > %s', r_stop, r_idx)
                break
            elif tag != datafile.TAG_COLLECTION_START:
                raise ValueError('invalid file format: not collection start')
            r_idx_next = r_idx + incr
            if tgt_r_idx >= r_idx_next:
                self._f.skip()
                r_idx = r_idx_next
                continue
            self._f.collection_goto_end()
            tag, value = next(self._f)
            if tag != datafile.TAG_COLLECTION_END:
                raise ValueError('invalid file format: not collection end')

            data = self._reduction_handler(value)
            self._reduction_cache = {
                'r_start': r_idx,
                'r_stop': r_idx_next,
                'buffer': data,
                'next_collection_pos': self._f.tell()
            }
            return self._reduction_cache

    def _reduction_handler_v1(self, value):
        field_count = len(self.config['reduction_fields'])
        b = np.frombuffer(value, dtype=np.float32).reshape((-1, field_count, _STATS_VALUES_V1))
        data = stats_array_factory(len(b))
        stats_array_invalidate(data)
        samples_per_reduction = self.config['samples_per_reduction']
        data[:, :]['length'] = samples_per_reduction
        for idx in range(field_count):
            data[:, idx]['mean'] = b[:, 0, 0]
            data[:, idx]['variance'] = (samples_per_reduction - 1) * b[:, 0, 1] * b[:, 0, 1]
            data[:, idx]['min'] = b[:, 0, 2]
            data[:, idx]['min'] = b[:, 0, 3]
        return data

    def _reduction_handler_v2(self, value):
        value = np.frombuffer(value, dtype=np.uint8)
        data, sample_count = array_storage.unpack(value)
        stats_array = stats_array_factory(sample_count)
        stats_array_invalidate(stats_array)
        for key, x in data.items():
            field_idx = (key >> 4) & 0x0f
            stats_idx = key & 0x0f
            stats_name = NP_STATS_NAMES[stats_idx]
            stats_array[:, field_idx][stats_name] = x  # .astype(dtype=np.float64)
        for idx in range(1, len(STATS_FIELD_NAMES)):
            stats_array[:, idx]['length'] = stats_array[:, 0]['length']
        return stats_array

    def get_reduction(self, start=None, stop=None, units=None, out=None):
        """Get the fixed reduction with statistics.

        :param start: The starting sample identifier (inclusive).
        :param stop: The ending sample identifier (exclusive).
        :param units: The units for start and stop.
            'seconds' or None is in floating point seconds relative to the view.
            'samples' is in stream buffer sample indices.
        :return: The The np.ndarray((N, STATS_FIELD_COUNT), dtype=DTYPE)
            reduction data which normally is memory mapped to the underlying
            data, but will be copied on rollover.
        """
        start, stop = self.normalize_time_arguments(start, stop, units)
        self._fh.seek(self._data_start_position)
        self._validate_range(start, stop)

        sz = self.config['samples_per_reduction']
        if sz < 1:
            sz = 1
        r_start = start // sz
        total_length = (stop - start) // sz
        r_stop = r_start + total_length
        log.info('DataReader.get_reduction(r_start=%r,r_stop=%r)', r_start, r_stop)

        if total_length <= 0:
            return stats_array_factory(0)
        if out is None:
            out = stats_array_factory(total_length)
        elif len(out) < total_length:
            raise ValueError('out too small')

        r_idx = r_start
        out_idx = 0

        while r_idx < r_stop:
            reduction_cache = self._reduction_tlv(r_idx)
            if reduction_cache is None:
                break
            data = reduction_cache['buffer']
            b_start = r_idx - reduction_cache['r_start']
            length = reduction_cache['r_stop'] - reduction_cache['r_start'] - b_start
            out_remaining = r_stop - r_idx
            length = min(length, out_remaining)
            if length <= 0:
                break
            out[out_idx:(out_idx + length), :] = data[b_start:(b_start + length), :]
            out_idx += length
            r_idx += length
        if out_idx != total_length:
            log.warning('DataReader length mismatch: out_idx=%s, length=%s', out_idx, total_length)
            total_length = min(out_idx, total_length)
        return out[:total_length, :]

    def _get_reduction_stats(self, start, stop):
        """Get statistics over the reduction

        :param start: The starting sample identifier (inclusive).
        :param stop: The ending sample identifier (exclusive).
        :return: The tuple of ((sample_start, sample_stop), :class:`Statistics`).
        """
        # log.debug('_get_reduction_stats(%s, %s)', start, stop)
        s = Statistics()
        sz = self.config['samples_per_reduction']
        incr = self.config['samples_per_block'] // sz
        r_start = start // sz
        if (r_start * sz) < start:
            r_start += 1
        r_stop = stop // sz
        if r_start >= r_stop:  # cannot use the reductions
            s_start = r_start * sz
            return (s_start, s_start), s
        r_idx = r_start

        while r_idx < r_stop:
            reduction_cache = self._reduction_tlv(r_idx)
            if reduction_cache is None:
                break
            data = reduction_cache['buffer']
            b_start = r_idx - reduction_cache['r_start']
            length = reduction_cache['r_stop'] - reduction_cache['r_start'] - b_start
            out_remaining = r_stop - r_idx
            length = min(length, out_remaining)
            if length <= 0:
                break
            r = reduction_downsample(data, b_start, b_start + length, length)
            s.combine(Statistics(stats=r[0, :]))
            r_idx += length
        return (r_start * sz, r_stop * sz), s

    def _data_get_handler_none(self, start, stop, out):
        return None

    def _data_get_handler_raw(self, start, stop, out):
        _, d_bits, d_cal = self._raw(start, stop)
        i, v = d_cal[:, 0], d_cal[:, 1]
        out[:, 0]['mean'] = i
        out[:, 1]['mean'] = v
        out[:, 2]['mean'] = i * v
        out[:, 3]['mean'] = np.bitwise_and(d_bits, 0x0f)
        out[:, 4]['mean'] = np.bitwise_and(np.right_shift(d_bits, 4), 0x01)
        out[:, 5]['mean'] = np.bitwise_and(np.right_shift(d_bits, 5), 0x01)
        out[:, :]['variance'] = 0.0  # zero variance, only one sample!
        out[:, :]['min'] = np.nan  # min
        out[:, :]['max'] = np.nan  # max

    def _data_get_handler_float32_v2(self, start, stop, out):
        k = self._downsampled(start, stop, STATS_FIELD_NAMES)
        out[:, 0]['mean'] = k['current']
        out[:, 1]['mean'] = k['voltage']
        out[:, 2]['mean'] = k['power']
        out[:, 3]['mean'] = k['current_range']
        out[:, 4]['mean'] = k['current_lsb']
        out[:, 5]['mean'] = k['voltage_lsb']
        out[:, :]['variance'] = 0.0  # zero variance, only one sample!
        out[:, :]['min'] = np.nan  # min
        out[:, :]['max'] = np.nan  # max

    def data_get(self, start, stop, increment=None, units=None, out=None):
        """Get the samples with statistics.

        :param start: The starting sample id (inclusive).
        :param stop: The ending sample id (exclusive).
        :param increment: The number of raw samples.
        :param units: The units for start and stop.
            'seconds' or None is in floating point seconds relative to the view.
            'samples' is in stream buffer sample indices.
        :param out: The optional output np.ndarray((N, STATS_FIELD_COUNT), dtype=DTYPE) to populate with
            the result.  None (default) will construct and return a new array.
        :return: The np.ndarray((N, STATS_FIELD_COUNT), dtype=DTYPE) data.
        """
        log.debug('DataReader.get(start=%r,stop=%r,increment=%r)', start, stop, increment)
        start, stop = self.normalize_time_arguments(start, stop, units)
        if increment is None:
            increment = 1
        if self._fh is None:
            raise IOError('file not open')
        increment = max(1, int(np.round(increment)))
        out_len = (stop - start) // increment
        if out_len <= 0:
            return stats_array_factory(0)
        if out is not None:
            if len(out) < len(out_len):
                raise ValueError('out too small')
        else:
            out = stats_array_factory(out_len)

        out_len = len(out)
        if increment == 1:
            self._data_get_handler(start, stop, out)
        elif increment == self.config['samples_per_reduction']:
            out = self.get_reduction(start, stop, out=out)
        elif increment > self.config['samples_per_reduction']:
            r_out = self.get_reduction(start, stop)
            increment = int(increment / self.config['samples_per_reduction'])
            out = reduction_downsample(r_out, 0, len(r_out), increment)
        else:
            k_start = start
            for idx in range(out_len):
                k_stop = k_start + increment
                self._statistics_get_handler(k_start, k_stop, out[idx, :])
                k_start = k_stop
        return out

    def summary_string(self):
        s = [str(self)]
        config_fields = ['sampling_frequency', 'samples_per_reduction', 'samples_per_tlv', 'samples_per_block']
        for field in config_fields:
            s.append('    %s = %r' % (field, self.config[field]))
        return '\n'.join(s)

    def time_to_sample_id(self, t):
        if t is None:
            return None
        s_min, s_max = self.sample_id_range
        s = int(t * self.sampling_frequency)
        if s < s_min or s > s_max:
            log.warning(f'time %.6f out of range', t)
            return None
        return s

    def sample_id_to_time(self, s):
        if s is None:
            return None
        return s / self.sampling_frequency

    def normalize_time_arguments(self, start, stop, units=None):
        """Normalize time arguments to range.

        :param start: The start time or samples.
            None gets the first sample, equivalent to self.sample_id_range[0].
        :param stop: The stop time or samples.
            None gets the last sample, equivalent to self.sample_id_range[1].
        :param units: The time units which is one of ['seconds', 'samples']
            None (default) id equivalent to 'samples'.
        :return: (start_sample, stop_sample).

        When units=='samples', negative values are interpreted in standard
        Python notation relative to the last sample.  None

        """
        s_min, s_max = self.sample_id_range
        if units == 'seconds':
            start = self.time_to_sample_id(start)
            stop = self.time_to_sample_id(stop)
        elif units is None or units == 'samples':
            if start is not None and start < 0:
                start = s_max + start
            if stop is not None and stop < 0:
                stop = s_max + stop
        else:
            raise ValueError(f'invalid time units: {units}')
        s1 = s_min if start is None else start
        s2 = s_max if stop is None else stop
        if s1 == s_min and s1 == s2:
            pass  # ok, zero length capture
        elif not s_min <= s1 < s_max:
            raise ValueError(f'start sample out of range: {s1}')
        elif not s_min <= s2 <= s_max:
            raise ValueError(f'start sample out of range: {s2}')
        return s1, s2

    def _stats_update(self, stats, x, length):
        stats['mean'] = np.mean(x, axis=0)
        stats['variance'] = np.var(x, axis=0) * length
        stats['min'] = np.amin(x, axis=0)
        stats['max'] = np.amax(x, axis=0)

    def _statistics_get_handler_none(self, start, stop, stats):
        return stop - start

    def _statistics_get_handler_raw(self, s1, s2, stats):
        _, d_bits, z = self._raw(s1, s2)
        i, v = z[:, 0], z[:, 1]
        p = i * v
        zi = np.isfinite(i)
        i_view = i[zi]
        length = len(i_view)
        stats[:]['length'] = length
        if length:
            i_range = np.bitwise_and(d_bits, 0x0f)
            i_lsb = np.right_shift(np.bitwise_and(d_bits, 0x10), 4)
            v_lsb = np.right_shift(np.bitwise_and(d_bits, 0x20), 5)
            for idx, field in enumerate([i_view, v[zi], p[zi], i_range, i_lsb[zi], v_lsb[zi]]):
                self._stats_update(stats[idx], field, length)
        else:
            stats_invalidate(stats)
        return length

    def _statistics_get_handler_float32_v2(self, s1, s2, stats):
        rv = {'signals': {}}
        self._samples_get_handler_float32_v2(s1, s2, STATS_FIELD_NAMES, rv)
        zi = np.isfinite(rv['signals']['current']['value'])
        length = np.count_nonzero(zi)
        stats[:]['length'] = length
        if length:
            for idx, field in enumerate(STATS_FIELD_NAMES):
                self._stats_update(stats[idx], rv['signals'][field]['value'][zi], length)
        else:
            stats_invalidate(stats)
        return length

    def _samples_get_handler_none(self, start, stop, fields, rv):
        return None

    def _samples_get_handler_raw(self, start, stop, fields, rv):
        raw, bits, cal = self._raw(start, stop)
        signals = rv['signals']
        if fields is None:
            fields = ['current', 'voltage', 'power', 'current_range', 'current_lsb', 'voltage_lsb', 'raw']
        for field in fields:
            d = {'units': _SIGNALS_UNITS.get(field, '')}
            if field == 'current':
                v = cal[:, 0]
            elif field == 'voltage':
                v = cal[:, 1]
            elif field == 'power':
                v = cal[:, 0] * cal[:, 1]
            elif field == 'raw':
                v = raw
                d['voltage_range'] = self._voltage_range
            elif field == 'raw_current':
                v = np.right_shift(raw[0::2], 2)
            elif field == 'raw_voltage':
                v = np.right_shift(raw[1::2], 2)
            elif field == 'bits':
                v = bits
            elif field == 'current_range':
                v = np.bitwise_and(bits, 0x0f)
            elif field == 'current_lsb':
                v = np.bitwise_and(np.right_shift(bits, 4), 1)
            elif field == 'voltage_lsb':
                v = np.bitwise_and(np.right_shift(bits, 5), 1)
            else:
                log.warning('Unsupported field %s', field)
                v = np.array([])
            d['value'] = v
            signals[field] = d
        return rv

    def _samples_get_handler_float32_v2(self, start, stop, fields, rv):
        k = self._downsampled(start, stop, fields)
        for field, value in k.items():
            units = _SIGNALS_UNITS.get(field, '')
            rv['signals'][field] = {'value': value, 'units': units}
        return rv

    def samples_get(self, start=None, stop=None, units=None, fields=None):
        """Get exact samples over a range.

        :param start: The starting time.
        :param stop: The ending time.
        :param units: The units for start and stop.
            'seconds' or None is in floating point seconds relative to the view.
            'samples' is in stream buffer sample indicies.
        :param fields: The fields to get.
        """
        log.debug('samples_get(%s, %s, %s)', start, stop, units)
        s1, s2 = self.normalize_time_arguments(start, stop, units)
        t1, t2 = s1 / self.sampling_frequency, s2 / self.sampling_frequency
        rv = {
            'time': {
                'range': {'value': [t1, t2], 'units': 's'},
                'delta':  {'value': t2 - t1, 'units': 's'},
                'sample_id_range':  {'value': [s1, s2], 'units': 'samples'},
                'sample_id_limits': {'value': self.sample_id_range, 'units': 'samples'},
                'samples': {'value': s2 - s1, 'units': 'samples'},
                'input_sampling_frequency': {'value': self.input_sampling_frequency, 'units': 'Hz'},
                'output_sampling_frequency': {'value': self.output_sampling_frequency, 'units': 'Hz'},
                'sampling_frequency': {'value': self.sampling_frequency, 'units': 'Hz'},
            },
            'signals': {},
        }
        return self._samples_get_handler(s1, s2, fields, rv)

    def _statistics_get(self, start, stop):
        s1, s2 = start, stop
        (k1, k2), s = self._get_reduction_stats(s1, s2)
        if k1 >= k2:
            # compute directly over samples
            stats = stats_factory()
            if not self._statistics_get_handler(s1, s2, stats):
                stats[3]['mean'] = I_RANGE_MISSING
                stats[3]['variance'] = 0
                stats[3]['min'] = I_RANGE_MISSING
                stats[3]['max'] = I_RANGE_MISSING
            s = Statistics(stats=stats)
        else:
            if s1 < k1:
                s_start = stats_factory()
                self._statistics_get_handler(s1, k1, s_start)
                s.combine(Statistics(stats=s_start))
            if s2 > k2:
                s_stop = stats_factory()
                self._statistics_get_handler(k2, s2, s_stop)
                s.combine(Statistics(stats=s_stop))
        return s

    def statistics_get(self, start=None, stop=None, units=None):
        """Get the statistics for the collected sample data over a time range.

        :param start: The starting time relative to the first sample.
        :param stop: The ending time.
        :param units: The units for start and stop.
            'seconds' is in floating point seconds relative to the view.
            'samples' or None is in stream buffer sample indices.
        :return: The statistics data structure.
            See the `statistics documentation <statistics.html>`_
            for details.
        """
        log.debug('statistics_get(%s, %s, %s)', start, stop, units)
        s1, s2 = self.normalize_time_arguments(start, stop, units)
        if s1 == s2:
            s2 = s1 + 1  # always try to produce valid statistics
        s = self._statistics_get(s1, s2)
        t_start = s1 / self.sampling_frequency
        t_stop = s2 / self.sampling_frequency
        return stats_to_api(s.value, t_start, t_stop)



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ctypes import windll, Structure, Union, \
    POINTER, pointer, byref, \
    c_void_p, c_wchar_p, memset, sizeof
from ctypes.wintypes import DWORD, HANDLE, BOOL, LPVOID, LPWSTR


_kernel32 = windll.kernel32


# CreateFile dwDesiredAccess values
GENERIC_WRITE = 1073741824
GENERIC_READ = -2147483648

# CreateFile dwShareMode flags
FILE_SHARE_READ = 1
FILE_SHARE_WRITE = 2
FILE_SHARE_DELETE = 4

# CreateFile dwCreationDisposition flags
CREATE_ALWAYS = 2
CREATE_NEW = 1
OPEN_ALWAYS = 4
OPEN_EXISTING = 3
TRUNCATE_EXISTING = 5

FILE_ATTRIBUTE_ARCHIVE = 0x20
FILE_ATTRIBUTE_ENCRYPTED = 0x4000
FILE_ATTRIBUTE_HIDDEN = 0x2
FILE_ATTRIBUTE_NORMAL = 0x80
FILE_ATTRIBUTE_OFFLINE = 0x1000
FILE_ATTRIBUTE_READONLY = 0x1
FILE_ATTRIBUTE_SYSTEM = 0x4
FILE_ATTRIBUTE_TEMPORARY = 0x100

FILE_FLAG_BACKUP_SEMANTICS = 0x02000000
FILE_FLAG_DELETE_ON_CLOSE = 0x04000000
FILE_FLAG_NO_BUFFERING = 0x20000000
FILE_FLAG_OPEN_NO_RECALL = 0x00100000
FILE_FLAG_OPEN_REPARSE_POINT = 0x00200000
FILE_FLAG_OVERLAPPED = 0x40000000
FILE_FLAG_POSIX_SEMANTICS = 0x0100000
FILE_FLAG_RANDOM_ACCESS = 0x10000000
FILE_FLAG_SESSION_AWARE = 0x00800000
FILE_FLAG_SEQUENTIAL_SCAN = 0x08000000
FILE_FLAG_WRITE_THROUGH = 0x80000000

INVALID_HANDLE_VALUE = HANDLE(-1)

# Common IO errors
ERROR_FILE_NOT_FOUND = 2  # The system cannot find the file specified.
ERROR_INVALID_HANDLE = 6  # Invalid handle
ERROR_BAD_COMMAND = 22  # The device does not recognize the command.
ERROR_GEN_FAILURE = 31  # A device attached to the system is not functioning.
ERROR_IO_OPERATION_ABORTED = 995
ERROR_IO_INCOMPLETE = 996
ERROR_IO_PENDING = 997
ERROR_DEVICE_NOT_CONNECTED = 1167

# NTSTATUS value
STATUS_TIMEOUT = 0x00000102
STATUS_PENDING = 0x00000103

WAIT_ABANDONED = 0x00000080
WAIT_OBJECT_0 = 0x00000000
WAIT_TIMEOUT = 0x00000102
WAIT_FAILED = 0xFFFFFFFF

MAXIMUM_WAIT_OBJECTS = 64


class _OverlappedOffset(Structure):
    _fields_ = [
        ('Offset', DWORD),
        ('OffsetHigh', DWORD),
    ]


class _OverlappedUnion(Union):
    _fields_ = [
        ('o', _OverlappedOffset),
        ('Pointer', LPVOID),
    ]


class Overlapped(Structure):
    _fields_ = [
        ('Internal', LPVOID),
        ('InternalHigh', LPVOID),
        ('u', _OverlappedUnion),
        ('hEvent', HANDLE),
    ]

    def __init__(self, hEvent):
        memset(pointer(self), 0, sizeof(self))
        self.hEvent = hEvent

    def reset(self):
        hEvent = self.hEvent
        memset(pointer(self), 0, sizeof(self))
        self.hEvent = hEvent


class SecurityAttributes(Structure):
    _fields_ = [
        ('n_length', DWORD),
        ('lp_security_descriptor', c_void_p),
        ('b_inherit_handle',BOOL)
    ]


# HANDLE WINAPI CreateFile(
#   _In_     LPCTSTR               lpFileName,
#   _In_     DWORD                 dwDesiredAccess,
#   _In_     DWORD                 dwShareMode,
#   _In_opt_ LPSECURITY_ATTRIBUTES lpSecurityAttributes,
#   _In_     DWORD                 dwCreationDisposition,
#   _In_     DWORD                 dwFlagsAndAttributes,
#   _In_opt_ HANDLE                hTemplateFile);
CreateFile = _kernel32.CreateFileW
CreateFile.restype = BOOL
CreateFile.argtypes = [c_wchar_p, DWORD, DWORD, POINTER(SecurityAttributes), DWORD, DWORD, HANDLE]

# BOOL WINAPI CloseHandle(_In_ HANDLE hObject);
CloseHandle = _kernel32.CloseHandle
CloseHandle.restype = BOOL
CloseHandle.argtypes = [HANDLE]

# BOOL WINAPI ReadFile(
#   _In_        HANDLE       hFile,
#   _Out_       LPVOID       lpBuffer,
#   _In_        DWORD        nNumberOfBytesToRead,
#   _Out_opt_   LPDWORD      lpNumberOfBytesRead,
#   _Inout_opt_ LPOVERLAPPED lpOverlapped);
ReadFile = _kernel32.ReadFile
ReadFile.restype = BOOL
ReadFile.argtypes = [HANDLE, c_void_p, DWORD, POINTER(DWORD), POINTER(Overlapped)]

# BOOL WINAPI WriteFile(
#   _In_        HANDLE       hFile,
#   _In_        LPCVOID      lpBuffer,
#   _In_        DWORD        nNumberOfBytesToWrite,
#   _Out_opt_   LPDWORD      lpNumberOfBytesWritten,
#   _Inout_opt_ LPOVERLAPPED lpOverlapped);
WriteFile = _kernel32.WriteFile
WriteFile.restype = BOOL
WriteFile.argtypes = [HANDLE, c_void_p, DWORD, POINTER(DWORD), POINTER(Overlapped)]

# BOOL WINAPI CancelIo(_In_ HANDLE hFile);
CancelIo = _kernel32.CancelIo
CancelIo.restype = BOOL
CancelIo.argtypes = [HANDLE]

#BOOL WINAPI CancelIoEx(
#  _In_     HANDLE       hFile,
#  _In_opt_ LPOVERLAPPED lpOverlapped);
CancelIoEx = _kernel32.CancelIoEx
CancelIoEx.restype = BOOL
CancelIoEx.argtypes = [HANDLE, POINTER(Overlapped)]

# HANDLE WINAPI CreateEvent(
#   _In_opt_ LPSECURITY_ATTRIBUTES lpEventAttributes,
#   _In_     BOOL                  bManualReset,
#   _In_     BOOL                  bInitialState,
#   _In_opt_ LPCTSTR               lpName);
CreateEvent = _kernel32.CreateEventW
CreateEvent.restype = HANDLE
CreateEvent.argtypes = [POINTER(SecurityAttributes), BOOL, BOOL, c_wchar_p]

# BOOL WINAPI SetEvent(_In_ HANDLE hEvent);
SetEvent = _kernel32.SetEvent
SetEvent.restype = BOOL
SetEvent.argtypes = [HANDLE]

# BOOL WINAPI ResetEvent(_In_ HANDLE hEvent);
ResetEvent = _kernel32.ResetEvent
ResetEvent.restype = BOOL
ResetEvent.argtypes = [HANDLE]

# DWORD WINAPI WaitForSingleObject(
#   _In_ HANDLE hHandle,
#   _In_ DWORD  dwMilliseconds);
WaitForSingleObject = _kernel32.WaitForSingleObject
WaitForSingleObject.restype = DWORD
WaitForSingleObject.argtypes = [HANDLE, DWORD]

# DWORD WaitForMultipleObjects(
#   DWORD        nCount,
#   const HANDLE *lpHandles,
#   BOOL         bWaitAll,
#   DWORD        dwMilliseconds);
WaitForMultipleObjects = _kernel32.WaitForMultipleObjects
WaitForMultipleObjects.restype = DWORD
WaitForMultipleObjects.argtypes = [DWORD, POINTER(HANDLE), BOOL, DWORD]

# DWORD WINAPI GetLastError(void);
GetLastError = _kernel32.GetLastError
GetLastError.restype = DWORD
GetLastError.argtypes = []

# FormatMessage dwFlags
FORMAT_MESSAGE_ALLOCATE_BUFFER = 0x00000100
FORMAT_MESSAGE_ARGUMENT_ARRAY = 0x00002000
FORMAT_MESSAGE_FROM_HMODULE = 0x00000800
FORMAT_MESSAGE_FROM_STRING = 0x00000400
FORMAT_MESSAGE_FROM_SYSTEM = 0x00001000
FORMAT_MESSAGE_IGNORE_INSERTS = 0x00000200
FORMAT_MESSAGE_MAX_WIDTH_MASK = 0x000000FF

# DWORD WINAPI FormatMessage(
#   _In_     DWORD   dwFlags,
#   _In_opt_ LPCVOID lpSource,
#   _In_     DWORD   dwMessageId,
#   _In_     DWORD   dwLanguageId,
#   _Out_    LPTSTR  lpBuffer,  -- With dwFlags ALLOCATE_BUFFER, actually POINTER to LPTSTR
#   _In_     DWORD   nSize,     -- 0 with dwFlags ALLOCATE_BUFFER
#   _In_opt_ va_list *Arguments);
FormatMessage = _kernel32.FormatMessageW
FormatMessage.restype = DWORD
FormatMessage.argtypes = [DWORD, c_void_p, DWORD, DWORD, POINTER(c_wchar_p), DWORD, c_void_p]


def get_error_str(error_code):
    if error_code == 0:
        return '[0] No error'
    bufptr = LPWSTR()
    FormatMessage(
        (FORMAT_MESSAGE_FROM_SYSTEM |
         FORMAT_MESSAGE_IGNORE_INSERTS |
         FORMAT_MESSAGE_ARGUMENT_ARRAY |
         FORMAT_MESSAGE_ALLOCATE_BUFFER),
        0,
        error_code,
        0,
        byref(bufptr),
        0,
        0)

    if bufptr.value is None:
        s = '(None)'
    else:
        s = bufptr.value.strip()
    return '[%d] %s' % (error_code, s)


def get_last_error():
    dLastError = GetLastError()
    return get_error_str(dLastError)



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the circular plot buffer with data reduction
"""

import unittest
from joulescope.calibration import Calibration
import monocypher
import numpy as np
import os

MYPATH = os.path.dirname(os.path.abspath(__file__))


with open(os.path.join(MYPATH, 'calibration_01.dat'), 'rb') as f:
    CAL1 = f.read()

CAL1_VK = [-1150.5926667302124, -2041.675208990041]
CAL1_VG = [0.0009468113102263853, 0.00029342128123136175]
CAL1_IK = [-2806.717811538373, -2822.6645675137756, -2820.88100055503, -2822.6200621623125,
           -2820.462616626318, -2820.5826040862453, -2818.676888822763, 0.0, float('nan')]
CAL1_1G = [0.0006828914450133605, 0.00016375303534373303, 1.4626174421742075e-05, 1.4937616926812478e-06,
           1.4970493483159016e-07, 1.498388811827842e-08, 1.4984542304875239e-09, 0.0, float('nan')]


PRIVATE_KEY = bytes(range(32))
PUBLIC_KEY = monocypher.compute_signing_public_key(PRIVATE_KEY)


class TestCalibrate(unittest.TestCase):

    def test_save_load_unsigned(self):
        c1 = Calibration()
        c1.current_offset = CAL1_IK[:7]
        c1.current_gain = CAL1_1G[:7]
        c1.voltage_offset = CAL1_VK
        c1.voltage_gain = CAL1_VG
        data = c1.save()
        c2 = Calibration().load(data, keys=[])
        self.assertFalse(c2.signed)
        np.testing.assert_allclose(CAL1_VK, c2.voltage_offset)
        np.testing.assert_allclose(CAL1_VG, c2.voltage_gain)
        np.testing.assert_allclose(CAL1_IK, c2.current_offset)
        np.testing.assert_allclose(CAL1_1G, c2.current_gain)

    def test_save_load_signed(self):
        c1 = Calibration()
        c1.current_offset = CAL1_IK[:7]
        c1.current_gain = CAL1_1G[:7]
        c1.voltage_offset = CAL1_VK
        c1.voltage_gain = CAL1_VG
        data = c1.save(PRIVATE_KEY)
        # with open(os.path.join(MYPATH, 'calibration_01.dat'), 'wb') as f:
        #     f.write(data)
        c2 = Calibration().load(data, keys=[PUBLIC_KEY])
        self.assertTrue(c2.signed)
        np.testing.assert_allclose(CAL1_VK, c2.voltage_offset)
        np.testing.assert_allclose(CAL1_VG, c2.voltage_gain)
        np.testing.assert_allclose(CAL1_IK, c2.current_offset)
        np.testing.assert_allclose(CAL1_1G, c2.current_gain)

    def test_cal1(self):
        c = Calibration().load(CAL1, keys=[PUBLIC_KEY])
        self.assertTrue(c.signed)
        np.testing.assert_allclose(CAL1_VK, c.voltage_offset)
        np.testing.assert_allclose(CAL1_VG, c.voltage_gain)
        np.testing.assert_allclose(CAL1_IK, c.current_offset)
        np.testing.assert_allclose(CAL1_1G, c.current_gain)

    def test_load_save_load(self):
        c1 = Calibration().load(CAL1, keys=[])
        self.assertFalse(c1.signed)
        data = c1.save(private_key=PRIVATE_KEY)
        c2 = Calibration().load(data, keys=[PUBLIC_KEY])
        np.testing.assert_allclose(CAL1_VK, c2.voltage_offset)
        np.testing.assert_allclose(CAL1_VG, c2.voltage_gain)
        np.testing.assert_allclose(CAL1_IK, c2.current_offset)
        np.testing.assert_allclose(CAL1_1G, c2.current_gain)



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The Python time standard uses POSIX (UNIX) time which is defined relative
to Jan 1, 1970.  Python uses floating point seconds to express time.
This module defines a much simpler 64-bit fixed point integer for
representing time which is much friendlier for microcontrollers.
The value is 34Q30 with the upper 34 bits
to represent whole seconds and the lower 30 bits to represent fractional
seconds.  A value of 2**30 (1 << 30) represents 1 second.  This
representation gives a resolution of 2 ** -30 (approximately 1 nanosecond)
and a range of +/- 2 ** 33 (approximately 272 years).  The value is
signed to allow for simple arithmetic on the time either as a fixed value
or as deltas.

For more details on a compatible C implementation, see
https://github.com/mliberty1/embc/tree/master/include/embc
"""


import datetime
import time as pytime


EPOCH = datetime.datetime(2018, 1, 1, tzinfo=datetime.timezone.utc).timestamp()


def timestamp_now():
    """Get the current timestamp.

    :return: The timestamp for the current time.
    """
    return seconds_to_timestamp(pytime.time())


def seconds_to_timestamp(time_s):
    """Convert python (POSIX) time in seconds to timestamp.

    :param time_s: The python time in float seconds.
    :return: The timestamp.
    """
    return int((time_s - EPOCH) * 2**30)


def timestamp_to_seconds(timestamp):
    """Convert timestamp to python (POSIX) time in seconds.

    :param timestamp: The timestamp.
    :return: The python time in float seconds.
    """
    return (timestamp / 2**30) + EPOCH



# Copyright 2019 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import random
import shutil


class FileReplace:
    """Safely replace an existing file.

    :param filename: The full path and file name.
    :param mode: The mode for open().
    :param encoding: The encoding for open().
    """

    def __init__(self, filename, mode=None, encoding=None):
        self._filename = os.path.abspath(filename)
        unique = '_%d_%d' % (os.getpid(), random.randint(0, 9999))
        name, ext = os.path.splitext(self._filename)
        self._mode = mode
        self._encoding = encoding

        # use filenames in same directory.  NamedTemporaryFile may be on different volume.
        self._filename_new = '%s_new_%s%s' % (name, unique, ext)
        self._filename_bak = '%s_bak_%s%s' % (name, unique, ext)
        self._f = None

    def open(self):
        if self._f is not None:
            self.close()
        if os.path.isfile(self._filename):
            shutil.copy(self._filename, self._filename_bak)
        self._f = open(self._filename_new, mode=self._mode, encoding=self._encoding)
        return self._f

    def replace(self, source):
        if os.path.isfile(self._filename):
            os.replace(source, self._filename)
        else:
            os.rename(source, self._filename)

    def close(self):
        try:
            if self._f:
                self._f.close()
                self._f = None
                if os.path.isfile(self._filename):
                    os.replace(self._filename_new, self._filename)
                else:
                    os.rename(self._filename_new, self._filename)
        finally:
            self._cleanup()

    def _cleanup(self):
        if os.path.isfile(self._filename_new):
            os.unlink(self._filename_new)
        if os.path.isfile(self._filename_bak):
            os.unlink(self._filename_bak)

    def revert(self):
        if os.path.isfile(self._filename):
            os.replace(self._filename_bak, self._filename)
        else:
            os.rename(self._filename_bak, self._filename)

    def __enter__(self):
        return self.open()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            if self._f:
                self._f.close()
                self._f = None
                # file should still be intact
        else:
            self.close()



# Copyright 2020 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import signal
import time
import logging
from joulescope.driver import scan_require_one


def parser_config(p):
    """Display Joulescope statistics."""
    p.add_argument('--compare',
                   action='store_true',
                   help='Compare sensor statistics to host stream_buffer statistics.')
    return on_cmd


def on_cmd(args):
    quit_ = False
    device = scan_require_one(name='Joulescope', config='off')

    def do_quit(*args, **kwargs):
        nonlocal quit_
        quit_ = 'quit from SIGINT'

    def statistics_cbk(s, indicator=None):
        indicator = '' if indicator is None else indicator
        t = s['time']['range']['value'][0]
        i = s['signals']['current']['µ']
        v = s['signals']['voltage']['µ']
        p = s['signals']['power']['µ']
        c = s['accumulators']['charge']
        e = s['accumulators']['energy']

        fmts = ['{x:.9f}', '{x:.3f}', '{x:.9f}', '{x:.9f}', '{x:.9f}']
        values = []
        for k, fmt in zip([i, v, p, c, e], fmts):
            value = fmt.format(x=k['value'])
            value = f'{value} {k["units"]}'
            values.append(value)
        ', '.join(values)
        print(f"{indicator}{t:.1f}: " + ', '.join(values))

    def on_stop(event=0, message=''):
        print(f'on_stop {event}: {message}')

    signal.signal(signal.SIGINT, do_quit)
    try:
        if args.compare:
            device.statistics_callback_register(lambda s: statistics_cbk(s, '< '), 'sensor')
            device.statistics_callback_register(lambda s: statistics_cbk(s, '> '), 'stream_buffer')
        else:
            device.statistics_callback_register(statistics_cbk, 'sensor')
        device.parameter_set('buffer_duration', 1)
        device.open()
        device.parameter_set('i_range', 'auto')
        device.parameter_set('v_range', '15V')
        if args.compare:
            device.parameter_set('source', 'raw')
            device.start(stop_fn=on_stop)
        while not quit_:
            device.status()
            time.sleep(0.100)
    except Exception as ex:
        logging.getLogger().exception('While getting statistics')
        print('Data streaming failed')
        return 1
    finally:
        device.close()
    return 0



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the JLS v2 writer.
"""

import unittest
from joulescope import JlsWriter
from joulescope.stream_buffer import StreamBuffer, usb_packet_factory
from pyjls import Reader
import tempfile
import numpy as np
import os
import shutil

PACKET_HEADER_SIZE = 8
SAMPLES_PER_PACKET = (512 - PACKET_HEADER_SIZE) // 4


class FakeJS110:

    def __init__(self):
        self._sampling_frequency = 2000000

    def __str__(self):
        return 'Joulescope:000001'

    def parameter_get(self, name):
        if name == 'sampling_frequency':
            return self._sampling_frequency
        else:
            raise ValueError(r'invalid parameter {name}')

    def parameter_set(self, name, value):
        if name == 'sampling_frequency':
            self._sampling_frequency = int(value)
        else:
            raise ValueError(r'invalid parameter {name}')

    def info(self):
        return {
            'type': 'info',
            'ver': 1,
            'ctl': {
                'mfg': {
                    'country': 'USA',
                    'location': 'MD_01',
                    'lot': '201927_00'},
                'hw': {
                    'rev': 'H',
                    'sn_mcu': '0123456789abcdef0123456789abcdef',
                    'sn_mfg': '000001'},
                'fw': {'ver': '1.3.2'},
                'fpga': {'ver': '0.2.0', 'prod_id': '0x9314acf2'}},
            'sensor': {
                'fw': {'ver': '1.3.2'},
                'fpga': {'ver': '1.2.1'}},
        }


class TestJlsWriter(unittest.TestCase):

    def setUp(self):
        self._tempdir = tempfile.mkdtemp()
        self._filename1 = os.path.join(self._tempdir, 'f1.jls')

    def tearDown(self):
        shutil.rmtree(self._tempdir)

    def test_empty_file(self):
        self.assertFalse(os.path.isfile(self._filename1))
        d = FakeJS110()
        with JlsWriter(d, self._filename1) as wr:
            pass  # empty file
        self.assertTrue(os.path.isfile(self._filename1))

    def _create_file(self, packet_index, count=None, signals=None):
        d = FakeJS110()
        stream_buffer = StreamBuffer(60.0, [10], 1000.0)
        stream_buffer.suppress_mode = 'off'
        if packet_index > 0:
            data = usb_packet_factory(0, packet_index - 1)
            stream_buffer.insert(data)
            stream_buffer.process()
        with JlsWriter(d, self._filename1, signals=signals) as wr:
            wr.stream_notify(stream_buffer)
            data = usb_packet_factory(packet_index, count)
            stream_buffer.insert(data)
            stream_buffer.process()
            wr.stream_notify(stream_buffer)
        self.assertTrue(os.path.isfile(self._filename1))

    def test_simple_file(self):
        self._create_file(0, 2)
        with Reader(self._filename1) as r:
            self.assertEqual([0, 1], sorted(r.sources.keys()))
            self.assertEqual(3, len(r.signals))
            self.assertEqual([0, 1, 2], sorted(r.signals.keys()))
            i = r.signals[1]
            self.assertEqual('current', i.name)
            self.assertEqual('A', i.units)
            self.assertEqual(252, i.length)
            v = r.signals[2]
            self.assertEqual('voltage', v.name)
            self.assertEqual('V', v.units)
            self.assertEqual(252, v.length)

            i_data = r.fsr(1, 0, i.length)
            np.testing.assert_allclose(np.arange(0, i.length * 2, 2), i_data)





# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from joulescope.data_recorder import DataReader
from joulescope.datafile import DataFileReader
import numpy as np
import sys


SUDO = "sudo " if "linux" in sys.platform else ""


MATPLOTLIB_IMPORT_ERROR = f"""
  _____                            _ 
 |  ___|                          | |
 | |__   _ __  _ __   ___   _ __  | |
 |  __| | '__|| '__| / _ \ | '__| | |
 | |___ | |   | |   | (_) || |    |_|
 \____/ |_|   |_|    \___/ |_|    (_)

Could not import matplotlib.

To install matplotlib, type:
    {SUDO}pip3 install -U matplotlib
then try running this program again.
"""


def parser_config(p):
    """Inspect recordings."""
    p.add_argument('filename',
                   help='The JLS filename to process.')
    p.add_argument('--plot-reduction',
                   action='store_true',
                   help='Plot the stored data reduction preview of the captured data.')
    p.add_argument('--plot',
                   action='store_true',
                   help='Plot the calibrated, captured data.')
    p.add_argument('--plot-raw',
                   help='Plot the raw data with list of plots.'
                        "'i'=current, 'v'=voltage, 'r'=current range. "
                        'Time range must < 2,000,000 samples')
    p.add_argument('--pretty_print',
                   action='store_true',
                   help='Print the file datastructure.')
    p.add_argument('--export',
                   help='Filename for export data.  ".csv" and ".npy" supported.')
    p.add_argument('--start',
                   default=0,
                   type=int,
                   help='Starting sample index.  Use 0 by default')
    p.add_argument('--stop',
                   default=-1,
                   type=int,
                   help='Stopping sample index.  Use last available by default')
    return on_cmd


def on_cmd(args):
    if args.plot_reduction or args.plot or args.plot_raw:
        try:
            import matplotlib.pyplot as plt
        except Exception:
            print(MATPLOTLIB_IMPORT_ERROR)
            return 1

    r = DataReader().open(args.filename)
    print(r.summary_string())
    start = args.start
    stop = args.stop
    if stop < 0:
        stop = r.sample_id_range[1] + 1 + stop

    if args.pretty_print:
        with open(args.filename, 'rb') as f:
            rf = DataFileReader(f)
            rf.pretty_print()

    if args.export is not None:
        k = r.samples_get(start, stop, units='samples', fields=['current', 'voltage'])
        i = k['signals']['current']['value']
        v = k['signals']['voltage']['value']
        data = np.hstack((i.reshape((-1, 1)), (v.reshape((-1, 1)))))
        if args.export.endswith('npy'):
            np.save(args.export, data)
        else:
            np.savetxt(args.export, data, fmt='%.5g', delimiter=',')

    if args.plot_reduction:
        y = r.get_reduction(start, stop)
        x = np.arange(len(y)) * (r.config['samples_per_reduction'] / r.config['sampling_frequency'])
        f = plt.figure()
        fields = r.config['reduction_fields']
        for axis, name in enumerate(fields):
            ax = f.add_subplot(len(fields), 1, axis + 1)
            ax.fill_between(x, y[:, axis]['min'], y[:, axis]['max'], color=(0.5, 0.5, 1.0, 0.5))
            ax.plot(x, y[:, axis]['mean'], color='blue')
            ax.set_ylabel(name)
            ax.grid(True)

        plt.show()
        plt.close(f)

    if args.plot:
        k = r.samples_get(start, stop, units='samples', fields=['current', 'voltage'])
        i = k['signals']['current']['value']
        v = k['signals']['voltage']['value']
        x = np.arange(len(i)) * (1.0 / r.config['sampling_frequency'])
        f = plt.figure()

        ax_i = f.add_subplot(2, 1, 1)
        ax_i.plot(x, i)
        ax_i.set_ylabel('Current (A)')
        ax_i.grid(True)

        ax_v = f.add_subplot(2, 1, 2, sharex=ax_i)
        ax_v.plot(x, v)
        ax_v.set_ylabel('Voltage (V)')
        ax_v.grid(True)

        ax_v.set_xlabel('Time (s)')

        plt.show()
        plt.close(f)

    if args.plot_raw:
        if stop - start > 2000000:
            print('Time range too long, cannot --plot-raw')
        else:
            plot_idx_total = len(args.plot_raw)
            link_axis = None
            plot_idx = 1
            rv = r.samples_get(start=start, stop=stop, units='samples', fields=['raw', 'current_range'])
            d_raw = rv['signals']['raw']['value']
            i_sel = rv['signals']['current_range']['value']
            i_raw = np.right_shift(d_raw[:, 0], 2)
            v_raw = np.right_shift(d_raw[:, 1], 2)
            x = np.arange(len(i_raw)) * (1.0 / r.config['sampling_frequency'])

            f = plt.figure()
            f.suptitle('Joulescope Raw Data')

            for c in args.plot_raw:
                if c == 'i':
                    ax = f.add_subplot(plot_idx_total, 1, plot_idx, sharex=link_axis)
                    ax.plot(x, i_raw)
                    ax.set_ylabel('Current (LSBs)')
                elif c == 'v':
                    ax = f.add_subplot(plot_idx_total, 1, plot_idx, sharex=link_axis)
                    ax.plot(x, v_raw)
                    ax.set_ylabel('Voltage (LSBs)')
                elif c == 'r':
                    ax = f.add_subplot(plot_idx_total, 1, plot_idx, sharex=link_axis)
                    ax.plot(x, i_sel)
                    ax.set_ylabel('Current Range')
                else:
                    raise ValueError('unsupported plot: %s' % c)

                ax.grid(True)
                if link_axis is None:
                    link_axis = ax
                plot_idx += 1

            # plt.tight_layout()
            ax.set_xlabel('Time (s)')
            plt.show()
            plt.close(f)

    r.close()
    return 0
